from typing import Union

from unicode_rbnf import RbnfEngine, FormatPurpose

from ovos_number_parser.numbers_az import numbers_to_digits_az, extract_number_az, is_fractional_az, pronounce_number_az
from ovos_number_parser.numbers_ca import numbers_to_digits_ca, pronounce_number_ca, is_fractional_ca, extract_number_ca
from ovos_number_parser.numbers_cs import numbers_to_digits_cs, pronounce_number_cs, is_fractional_cs, extract_number_cs
from ovos_number_parser.numbers_da import numbers_to_digits_da, pronounce_number_da, is_fractional_da, is_ordinal_da, \
    pronounce_ordinal_da, extract_number_da
from ovos_number_parser.numbers_de import numbers_to_digits_de, pronounce_number_de, pronounce_ordinal_de, \
    is_ordinal_de, is_fractional_de, extract_number_de
from ovos_number_parser.numbers_en import numbers_to_digits_en, is_ordinal_en, pronounce_number_en, extract_number_en, \
    is_fractional_en
from ovos_number_parser.numbers_es import numbers_to_digits_es, pronounce_number_es, extract_number_es, is_fractional_es
from ovos_number_parser.numbers_eu import pronounce_number_eu, extract_number_eu, is_fractional_eu
from ovos_number_parser.numbers_fa import pronounce_number_fa, extract_number_fa
from ovos_number_parser.numbers_fr import (pronounce_number_fr, extract_number_fr, is_fractional_fr)
from ovos_number_parser.numbers_gl import pronounce_number_gl, extract_number_gl, is_fractional_gl, numbers_to_digits_gl
from ovos_number_parser.numbers_hu import pronounce_number_hu, pronounce_ordinal_hu
from ovos_number_parser.numbers_it import (extract_number_it, pronounce_number_it, is_fractional_it)
from ovos_number_parser.numbers_nl import numbers_to_digits_nl, pronounce_number_nl, pronounce_ordinal_nl, \
    extract_number_nl, is_fractional_nl
from ovos_number_parser.numbers_pl import numbers_to_digits_pl, pronounce_number_pl, extract_number_pl, is_fractional_pl
from ovos_number_parser.numbers_pt import PortugueseVariant, pronounce_fraction_pt, numbers_to_digits_pt, \
    pronounce_number_pt, is_fractional_pt, extract_number_pt, pronounce_ordinal_pt, is_ordinal_pt
from ovos_number_parser.numbers_ru import numbers_to_digits_ru, pronounce_number_ru, extract_number_ru, is_fractional_ru
from ovos_number_parser.numbers_sl import pronounce_number_sl
from ovos_number_parser.numbers_sv import pronounce_number_sv, pronounce_ordinal_sv, extract_number_sv, \
    is_fractional_sv
from ovos_number_parser.numbers_uk import numbers_to_digits_uk, pronounce_number_uk, extract_number_uk, is_fractional_uk
from ovos_number_parser.util import Scale, GrammaticalGender, DigitPronunciation


def numbers_to_digits(utterance: str, lang: str, scale: Scale = Scale.LONG) -> str:
    """
    Convert written numbers in a text string to their digit representations for the specified language and numerical scale.
    
    Parameters:
        utterance (str): Text potentially containing written numbers.
        lang (str): Language code used to determine parsing rules.
        scale (Scale, optional): Numerical scale (long or short) for languages that distinguish between them.
    
    Returns:
        str: The input text with written numbers replaced by their digit equivalents.
    
    Raises:
        NotImplementedError: If the specified language is not supported.
    """
    if lang.startswith("az"):
        return numbers_to_digits_az(utterance)
    if lang.startswith("ca"):
        return numbers_to_digits_ca(utterance)
    if lang.startswith("gl"):
        return numbers_to_digits_gl(utterance)
    if lang.startswith("cs"):
        return numbers_to_digits_cs(utterance)
    if lang.startswith("da"):
        return numbers_to_digits_da(utterance)
    if lang.startswith("de"):
        return numbers_to_digits_de(utterance)
    if lang.startswith("en"):
        return numbers_to_digits_en(utterance)
    if lang.startswith("es"):
        return numbers_to_digits_es(utterance)
    if lang.startswith("nl"):
        return numbers_to_digits_nl(utterance)
    if lang.startswith("pl"):
        return numbers_to_digits_pl(utterance)
    if lang.startswith("pt"):
        return numbers_to_digits_pt(utterance, scale=scale)
    if lang.startswith("ru"):
        return numbers_to_digits_ru(utterance)
    if lang.startswith("uk"):
        return numbers_to_digits_uk(utterance)
    raise NotImplementedError(f"Unsupported language: '{lang}'")


def pronounce_number(number: Union[int, float], lang: str,
                     places: int = 3,
                     short_scale: bool = True,
                     scientific: bool = False, ordinals: bool = False,
                     digits: DigitPronunciation = DigitPronunciation.FULL_NUMBER,
                     gender: GrammaticalGender = GrammaticalGender.MASCULINE) -> str:
    """
    Return the spoken representation of a number in the specified language.
     
    Converts a numeric value to its pronounced form, supporting various languages, decimal precision, scale (short or long), scientific notation, ordinal forms, digit pronunciation styles, and grammatical gender where applicable. Falls back to Unicode RBNF for unsupported languages.
     
    Parameters:
        number (int or float): The number to pronounce.
        lang (str): BCP-47 language code specifying the language for pronunciation.
        places (int, optional): Number of decimal places to include (default is 3).
        short_scale (bool, optional): Whether to use the short scale for large numbers (default is True).
        scientific (bool, optional): If True, pronounce the number in scientific notation.
        ordinals (bool, optional): If True, pronounce the number as an ordinal (e.g., "first" instead of "one").
        digits (DigitPronunciation, optional): Style for pronouncing digits (default is FULL_NUMBER).
        gender (GrammaticalGender, optional): Grammatical gender for languages that require it (default is MASCULINE).
     
    Returns:
        str: The pronounced form of the number.
     
    Raises:
        NotImplementedError: If the specified language is not supported.
    """
    scale = Scale.SHORT if short_scale else Scale.LONG  # TODO migrate function kwarg to accept Scale enum
    if lang.startswith("en"):
        return pronounce_number_en(number, places, short_scale, scientific, ordinals)
    if lang.startswith("az"):
        return pronounce_number_az(number, places, short_scale, scientific, ordinals)
    if lang.startswith("ca"):
        return pronounce_number_ca(number, places)
    if lang.startswith("cs"):
        return pronounce_number_en(number, places, short_scale, scientific, ordinals)
    if lang.startswith("da"):
        return pronounce_number_da(number, places, short_scale, scientific, ordinals)
    if lang.startswith("de"):
        return pronounce_number_de(number, places, short_scale, scientific, ordinals)
    if lang.startswith("gl"):
        return pronounce_number_gl(number, places)
    if lang.startswith("es"):
        return pronounce_number_es(number, places)
    if lang.startswith("eu"):
        return pronounce_number_eu(number, places)
    if lang.startswith("fa"):
        return pronounce_number_fa(number, places, scientific, ordinals)
    if lang.startswith("fr"):
        return pronounce_number_fr(number, places)
    if lang.startswith("hu"):
        return pronounce_number_hu(number, places, short_scale, scientific, ordinals)
    if lang.startswith("it"):
        return pronounce_number_it(number, places, short_scale, scientific)
    if lang.startswith("nl"):
        return pronounce_number_nl(number, places, short_scale, scientific, ordinals)
    if lang.startswith("pl"):
        return pronounce_number_pl(number, places, short_scale, scientific, ordinals)
    if lang.startswith("pt"):
        variant = PortugueseVariant.BR if "br" in lang.lower() else PortugueseVariant.PT
        return pronounce_number_pt(number, places, scale=scale,
                                   variant=variant, ordinals=ordinals,
                                   digits=digits, gender=gender)
    if lang.startswith("ru"):
        return pronounce_number_ru(number, places, short_scale, scientific, ordinals)
    if lang.startswith("sl"):
        return pronounce_number_sl(number, places, short_scale, scientific, ordinals)
    if lang.startswith("sv"):
        return pronounce_number_sv(number, places, short_scale, scientific, ordinals)
    if lang.startswith("uk"):
        return pronounce_number_uk(number, places, short_scale, scientific, ordinals)
    # fallback to unicode RBNF
    try:
        engine = RbnfEngine.for_language(lang.split("-")[0])
        fmt = FormatPurpose.ORDINAL if ordinals else FormatPurpose.CARDINAL
        return engine.format_number(number, fmt).text
    except Exception as err:
        raise NotImplementedError(f"Unsupported language: '{lang}'") from err


def pronounce_fraction(fraction_word: str, lang: str, scale: Scale = Scale.LONG) -> str:
    """
    Return the spoken form of a fraction string (e.g., "1/2" as "one half") for the specified language and numerical scale.
    
    Parameters:
        fraction_word (str): The fraction to pronounce (e.g., "3/2").
        scale (Scale, optional): Numerical scale to use (SHORT or LONG). Defaults to LONG.
    
    Returns:
        str: The pronounced fraction.
    
    Raises:
        NotImplementedError: If the specified language is not supported.
    """
    if lang.startswith("pt"):
        variant = PortugueseVariant.BR if "br" in lang.lower() else PortugueseVariant.PT
        return pronounce_fraction_pt(fraction_word, scale=scale, variant=variant)
    else:
        raise NotImplementedError(f"unsupported language: {lang}")


def pronounce_ordinal(number: Union[int, float], lang: str,
                      short_scale: bool = True,
                      gender: GrammaticalGender = GrammaticalGender.MASCULINE) -> str:
    """
    Return the spoken ordinal form of a number in the specified language.
      
    Parameters:
        number (int or float): The number to convert to its ordinal spoken equivalent.
        lang (str): BCP-47 language code specifying the language for pronunciation.
        short_scale (bool, optional): Whether to use the short (True) or long (False) scale for large numbers. Defaults to True.
        gender (GrammaticalGender, optional): Grammatical gender to use for languages that require it. Defaults to masculine.
      
    Returns:
        str: The ordinal number pronounced in the specified language.
      
    Raises:
        NotImplementedError: If the language is not supported.
    """
    scale = Scale.SHORT if short_scale else Scale.LONG  # TODO migrate function kwarg to accept Scale enum
    if lang.startswith("pt"):
        variant = PortugueseVariant.BR if "br" in lang.lower() else PortugueseVariant.PT
        return pronounce_ordinal_pt(number, scale=scale, variant=variant, gender=gender)
    if lang.startswith("da"):
        return pronounce_ordinal_da(number)
    if lang.startswith("de"):
        return pronounce_ordinal_de(number)
    if lang.startswith("hu"):
        return pronounce_ordinal_hu(number)
    if lang.startswith("nl"):
        return pronounce_ordinal_nl(number)
    if lang.startswith("sv"):
        return pronounce_ordinal_sv(number)
    # fallback to unicode RBNF
    try:
        engine = RbnfEngine.for_language(lang.split("-")[0])
        fmt = FormatPurpose.ORDINAL
        return engine.format_number(number, fmt).text
    except Exception as err:
        raise NotImplementedError(f"Unsupported language: '{lang}'") from err


def extract_number(text: str, lang: str, short_scale: bool = True, ordinals: bool = False) -> Union[int, float, bool]:
    """Takes in a string and extracts a number.

    Assumes only 1 number is in the string, does NOT handle multiple numbers

    Args:
        text (str): the string to extract a number from
        short_scale (bool): Use "short scale" or "long scale" for large
            numbers -- over a million.  The default is short scale, which
            is now common in most English speaking countries.
            See https://en.wikipedia.org/wiki/Names_of_large_numbers
        ordinals (bool): consider ordinal numbers, e.g. third=3 instead of 1/3
        lang (str, optional): an optional BCP-47 language code, if omitted
                              the default language will be used.
    Returns:
        (int, float or False): The number extracted or False if the input
                               text contains no numbers
    """
    scale = Scale.SHORT if short_scale else Scale.LONG  # TODO migrate function kwarg to accept Scale enum
    if lang.startswith("en"):
        return extract_number_en(text, short_scale, ordinals)
    if lang.startswith("az"):
        return extract_number_az(text, short_scale, ordinals)
    if lang.startswith("ca"):
        return extract_number_ca(text, short_scale, ordinals)
    if lang.startswith("cs"):
        return extract_number_cs(text, short_scale, ordinals)
    if lang.startswith("da"):
        return extract_number_da(text, short_scale, ordinals)
    if lang.startswith("de"):
        return extract_number_de(text, short_scale, ordinals)
    if lang.startswith("es"):
        return extract_number_es(text, short_scale, ordinals)
    if lang.startswith("gl"):
        return extract_number_gl(text, short_scale, ordinals)
    if lang.startswith("eu"):
        return extract_number_eu(text, short_scale, ordinals)
    if lang.startswith("fa"):
        return extract_number_fa(text, ordinals)
    if lang.startswith("fr"):
        return extract_number_fr(text, short_scale, ordinals)
    if lang.startswith("it"):
        return extract_number_it(text, short_scale, ordinals)
    if lang.startswith("nl"):
        return extract_number_nl(text, short_scale, ordinals)
    if lang.startswith("pl"):
        return extract_number_pl(text, short_scale, ordinals)
    if lang.startswith("pt"):
        variant = PortugueseVariant.BR if "br" in lang.lower() else PortugueseVariant.PT
        return extract_number_pt(text, scale=scale, ordinals=ordinals, variant=variant)
    if lang.startswith("ru"):
        return extract_number_ru(text, short_scale, ordinals)
    if lang.startswith("sv"):
        return extract_number_sv(text, short_scale, ordinals)
    if lang.startswith("uk"):
        return extract_number_uk(text, short_scale, ordinals)
    raise NotImplementedError(f"Unsupported language: '{lang}'")


def is_fractional(input_str: str, lang: str, short_scale: bool = True) -> Union[bool, float]:
    """
    This function takes the given text and checks if it is a fraction.
    Used by most of the number exractors.

    Will return False on phrases that *contain* a fraction. Only detects
    exact matches. To pull a fraction from a string, see extract_number()

    Args:
        input_str (str): the string to check if fractional
        short_scale (bool): use short scale if True, long scale if False
        lang (str, optional): an optional BCP-47 language code, if omitted
                              the default language will be used.
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction
    """
    if lang.startswith("en"):
        return is_fractional_en(input_str, short_scale)
    if lang.startswith("az"):
        return is_fractional_az(input_str, short_scale)
    if lang.startswith("ca"):
        return is_fractional_ca(input_str, short_scale)
    if lang.startswith("cs"):
        return is_fractional_cs(input_str, short_scale)
    if lang.startswith("da"):
        return is_fractional_da(input_str, short_scale)
    if lang.startswith("de"):
        return is_fractional_de(input_str, short_scale)
    if lang.startswith("es"):
        return is_fractional_es(input_str, short_scale)
    if lang.startswith("gl"):
        return is_fractional_gl(input_str, short_scale)
    if lang.startswith("eu"):
        return is_fractional_eu(input_str)
    if lang.startswith("fr"):
        return is_fractional_fr(input_str)
    if lang.startswith("it"):
        return is_fractional_it(input_str, short_scale)
    if lang.startswith("nl"):
        return is_fractional_pl(input_str, short_scale)
    if lang.startswith("pl"):
        return is_fractional_pl(input_str, short_scale)
    if lang.startswith("pt"):
        return is_fractional_pt(input_str)
    if lang.startswith("ru"):
        return is_fractional_ru(input_str, short_scale)
    if lang.startswith("sv"):
        return is_fractional_sv(input_str, short_scale)
    if lang.startswith("uk"):
        return is_fractional_uk(input_str, short_scale)
    raise NotImplementedError(f"Unsupported languags: '{lang}'")


def is_ordinal(input_str: str, lang: str) -> Union[bool, float]:
    """
    This function takes the given text and checks if it is an ordinal number.

    Args:
        input_str (str): the string to check if ordinal
        lang (str, optional): an optional BCP-47 language code, if omitted
                              the default language will be used.
    Returns:
        (bool) or (float): False if not an ordinal, otherwise the number
        corresponding to the ordinal
    """
    if lang.startswith("pt"):
        return is_ordinal_pt(input_str)
    if lang.startswith("en"):
        return is_ordinal_en(input_str)
    if lang.startswith("de"):
        return is_ordinal_de(input_str)
    if lang.startswith("da"):
        return is_ordinal_da(input_str)
    raise NotImplementedError(f"Unsupported languags: '{lang}'")



import collections

from ovos_number_parser.util import convert_to_mixed_fraction, is_numeric, look_for_fractions

_SHORT_ORDINAL_STRING_IT = {
    1: 'primo',
    2: 'secondo',
    3: 'terzo',
    4: 'quarto',
    5: 'quinto',
    6: 'sesto',
    7: 'settimo',
    8: 'ottavo',
    9: 'nono',
    10: 'decimo',
    11: 'undicesimo',
    12: 'dodicesimo',
    13: 'tredicesimo',
    14: 'quattordicesimo',
    15: 'quindicesimo',
    16: 'sedicesimo',
    17: 'diciassettesimo',
    18: 'diciottesimo',
    19: 'diciannovesimo',
    20: 'ventesimo',
    30: 'trentesimo',
    40: 'quarantesimo',
    50: 'cinquantesimo',
    60: 'sessantesimo',
    70: 'settantesimo',
    80: 'ottantesimo',
    90: 'novantesimo',
    1e2: 'centesimo',
    1e3: 'millesimo',
    1e6: 'milionesimo',
    1e9: 'miliardesimo',
    1e12: 'trilionesimo',
    1e15: 'quadrilionesimo',
    1e18: 'quintilionesim',
    1e21: 'sestilionesimo',
    1e24: 'settilionesimo',
    1e27: 'ottilionesimo',
    1e30: 'nonilionesimo',
    1e33: 'decilionesimo'
    # TODO > 1e-33
}

#  per i > 10e12 modificata solo la desinenza: da sistemare a fine debug
_LONG_ORDINAL_STRING_IT = {
    1: 'primo',
    2: 'secondo',
    3: 'terzo',
    4: 'quarto',
    5: 'quinto',
    6: 'sesto',
    7: 'settimo',
    8: 'ottavo',
    9: 'nono',
    10: 'decimo',
    11: 'undicesimo',
    12: 'dodicesimo',
    13: 'tredicesimo',
    14: 'quattordicesimo',
    15: 'quindicesimo',
    16: 'sedicesimo',
    17: 'diciassettesimo',
    18: 'diciottesimo',
    19: 'diciannovesimo',
    20: 'ventesimo',
    30: 'trentesimo',
    40: 'quarantesimo',
    50: 'cinquantesimo',
    60: 'sessantesimo',
    70: 'settantesimo',
    80: 'ottantesimo',
    90: 'novantesimo',
    1e2: 'centesimo',
    1e3: 'millesimo',
    1e6: 'milionesimo',
    1e12: 'bilionesimo',
    1e18: 'trilionesimo',
    1e24: 'quadrilionesimo',
    1e30: 'quintilionesimo',
    1e36: 'sestilionesimo',
    1e42: 'settilionesimo',
    1e48: 'ottilionesimo',
    1e54: 'nonilionesimo',
    1e60: 'decilionesimo'
    # TODO > 1e60
}

# Undefined articles ['un', 'una', 'un\''] can not be supressed,
# in Italian, 'un cavallo' means 'a horse' or 'one horse'.
_ARTICLES_IT = ['il', 'lo', 'la', 'i', 'gli', 'le']

_STRING_NUM_IT = {
    'zero': 0,
    'un': 1,
    'uno': 1,
    'una': 1,
    'un\'': 1,
    'due': 2,
    'tre': 3,
    'quattro': 4,
    'cinque': 5,
    'sei': 6,
    'sette': 7,
    'otto': 8,
    'nove': 9,
    'dieci': 10,
    'undici': 11,
    'dodici': 12,
    'tredici': 13,
    'quattordici': 14,
    'quindici': 15,
    'sedici': 16,
    'diciassette': 17,
    'diciotto': 18,
    'diciannove': 19,
    'venti': 20,
    'vent': 20,
    'trenta': 30,
    'trent': 30,
    'quaranta': 40,
    'quarant': 40,
    'cinquanta': 50,
    'cinquant': 50,
    'sessanta': 60,
    'sessant': 60,
    'settanta': 70,
    'settant': 70,
    'ottanta': 80,
    'ottant': 80,
    'novanta': 90,
    'novant': 90,
    'cento': 100,
    'duecento': 200,
    'trecento': 300,
    'quattrocento': 400,
    'cinquecento': 500,
    'seicento': 600,
    'settecento': 700,
    'ottocento': 800,
    'novecento': 900,
    'mille': 1000,
    'mila': 1000,
    'centomila': 100000,
    'milione': 1000000,
    'miliardo': 1000000000,
    'primo': 1,
    'secondo': 2,
    'mezzo': 0.5,
    'mezza': 0.5,
    'paio': 2,
    'decina': 10,
    'decine': 10,
    'dozzina': 12,
    'dozzine': 12,
    'centinaio': 100,
    'centinaia': 100,
    'migliaio': 1000,
    'migliaia': 1000
}

_NUM_STRING_IT = {
    0: 'zero',
    1: 'uno',
    2: 'due',
    3: 'tre',
    4: 'quattro',
    5: 'cinque',
    6: 'sei',
    7: 'sette',
    8: 'otto',
    9: 'nove',
    10: 'dieci',
    11: 'undici',
    12: 'dodici',
    13: 'tredici',
    14: 'quattordici',
    15: 'quindici',
    16: 'sedici',
    17: 'diciassette',
    18: 'diciotto',
    19: 'diciannove',
    20: 'venti',
    30: 'trenta',
    40: 'quaranta',
    50: 'cinquanta',
    60: 'sessanta',
    70: 'settanta',
    80: 'ottanta',
    90: 'novanta'
}

_FRACTION_STRING_IT = {
    2: 'mezz',
    3: 'terz',
    4: 'quart',
    5: 'quint',
    6: 'sest',
    7: 'settim',
    8: 'ottav',
    9: 'non',
    10: 'decim',
    11: 'undicesim',
    12: 'dodicesim',
    13: 'tredicesim',
    14: 'quattordicesim',
    15: 'quindicesim',
    16: 'sedicesim',
    17: 'diciassettesim',
    18: 'diciottesim',
    19: 'diciannovesim',
    20: 'ventesim'
}

# fonte: http://tulengua.es/numeros-texto/default.aspx
_LONG_SCALE_IT = collections.OrderedDict([
    (100, 'cento'),
    (1000, 'mila'),
    (1000000, 'milioni'),
    (1e9, "miliardi"),
    (1e12, "bilioni"),
    (1e18, 'trilioni'),
    (1e24, "quadrilioni"),
    (1e30, "quintilioni"),
    (1e36, "sestilioni"),
    (1e42, "settilioni"),
    (1e48, "ottillioni"),
    (1e54, "nonillioni"),
    (1e60, "decemillioni"),
    (1e66, "undicilione"),
    (1e72, "dodicilione"),
    (1e78, "tredicilione"),
    (1e84, "quattordicilione"),
    (1e90, "quindicilione"),
    (1e96, "sedicilione"),
    (1e102, "diciasettilione"),
    (1e108, "diciottilione"),
    (1e114, "dicianovilione"),
    (1e120, "vintilione"),
    (1e306, "unquinquagintilione"),
    (1e312, "duoquinquagintilione"),
    (1e336, "sesquinquagintilione"),
    (1e366, "unsexagintilione")
])

_SHORT_SCALE_IT = collections.OrderedDict([
    (100, 'cento'),
    (1000, 'mila'),
    (1000000, 'milioni'),
    (1e9, "miliardi"),
    (1e12, 'bilioni'),
    (1e15, "biliardi"),
    (1e18, "trilioni"),
    (1e21, "triliardi"),
    (1e24, "quadrilioni"),
    (1e27, "quadriliardi"),
    (1e30, "quintilioni"),
    (1e33, "quintiliardi"),
    (1e36, "sestilioni"),
    (1e39, "sestiliardi"),
    (1e42, "settilioni"),
    (1e45, "settiliardi"),
    (1e48, "ottilioni"),
    (1e51, "ottiliardi"),
    (1e54, "nonilioni"),
    (1e57, "noniliardi"),
    (1e60, "decilioni"),
    (1e63, "deciliardi"),
    (1e66, "undicilioni"),
    (1e69, "undiciliardi"),
    (1e72, "dodicilioni"),
    (1e75, "dodiciliardi"),
    (1e78, "tredicilioni"),
    (1e81, "trediciliardi"),
    (1e84, "quattordicilioni"),
    (1e87, "quattordiciliardi"),
    (1e90, "quindicilioni"),
    (1e93, "quindiciliardi"),
    (1e96, "sedicilioni"),
    (1e99, "sediciliardi"),
    (1e102, "diciassettilioni"),
    (1e105, "diciassettiliardi"),
    (1e108, "diciottilioni"),
    (1e111, "diciottiliardi"),
    (1e114, "dicianovilioni"),
    (1e117, "dicianoviliardi"),
    (1e120, "vintilioni"),
    (1e123, "vintiliardi"),
    (1e153, "quinquagintillion"),
    (1e183, "sexagintillion"),
    (1e213, "septuagintillion"),
    (1e243, "ottogintilioni"),
    (1e273, "nonigintillioni"),
    (1e303, "centilioni"),
    (1e306, "uncentilioni"),
    (1e309, "duocentilioni"),
    (1e312, "trecentilioni"),
    (1e333, "decicentilioni"),
    (1e336, "undicicentilioni"),
    (1e363, "viginticentilioni"),
    (1e366, "unviginticentilioni"),
    (1e393, "trigintacentilioni"),
    (1e423, "quadragintacentillion"),
    (1e453, "quinquagintacentillion"),
    (1e483, "sexagintacentillion"),
    (1e513, "septuagintacentillion"),
    (1e543, "ctogintacentillion"),
    (1e573, "nonagintacentillion"),
    (1e603, "ducentillion"),
    (1e903, "trecentillion"),
    (1e1203, "quadringentillion"),
    (1e1503, "quingentillion"),
    (1e1803, "sescentillion"),
    (1e2103, "septingentillion"),
    (1e2403, "octingentillion"),
    (1e2703, "nongentillion"),
    (1e3003, "millinillion")
])


def is_fractional_it(input_str, short_scale=False):
    """
    This function takes the given text and checks if it is a fraction.
    Updated to italian from en version 18.8.9

    Args:
        input_str (str): the string to check if fractional
        short_scale (bool): use short scale if True, long scale if False
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction

    """
    input_str = input_str.lower()
    if input_str.endswith('i', -1) and len(input_str) > 2:
        input_str = input_str[:-1] + "o"  # normalizza plurali

    fracts_it = {"intero": 1, "mezza": 2, "mezzo": 2}

    if short_scale:
        for num in _SHORT_ORDINAL_STRING_IT:
            if num > 2:
                fracts_it[_SHORT_ORDINAL_STRING_IT[num]] = num
    else:
        for num in _LONG_ORDINAL_STRING_IT:
            if num > 2:
                fracts_it[_LONG_ORDINAL_STRING_IT[num]] = num

    if input_str in fracts_it:
        return 1.0 / fracts_it[input_str]
    return False


def _extract_number_long_it(word):
    """
     This function converts a long textual number like
     milleventisette -> 1027 diecimila -> 10041 in
     integer value, covers from  0 to 999999999999999
     for now limited to 999_e21 but ready for 999_e63
     example:
        milleventisette -> 1027
        diecimilaquarantuno-> 10041
        centottomiladuecentotredici -> 108213
    Args:
         word (str): the word to convert in number
    Returns:
         (bool) or (int): The extracted number or False if no number
                                   was found
    """

    units = {'zero': 0, 'uno': 1, 'due': 2, 'tre': 3, 'quattro': 4,
             'cinque': 5, 'sei': 6, 'sette': 7, 'otto': 8, 'nove': 9}

    tens = {'dieci': 10, 'venti': 20, 'trenta': 30, 'quaranta': 40,
            'cinquanta': 50, 'sessanta': 60, 'settanta': 70, 'ottanta': 80,
            'novanta': 90}

    tens_short = {'vent': 20, 'trent': 30, 'quarant': 40, 'cinquant': 50,
                  'sessant': 60, 'settant': 70, 'ottant': 80, 'novant': 90}

    nums_long = {'undici': 11, 'dodici': 12, 'tredici': 13, 'quattordici': 14,
                 'quindici': 15, 'sedici': 16, 'diciassette': 17,
                 'diciotto': 18, 'diciannove': 19}

    multipli_it = collections.OrderedDict([
        # (1e63, 'deciliardi'),
        # (1e60, 'decilioni'),
        # (1e57, 'noviliardi'),
        # (1e54, 'novilioni'),
        # (1e51, 'ottiliardi'),
        # (1e48, 'ottilioni'),
        # (1e45, 'settiliardi'),
        # (1e42, 'settilioni'),
        # (1e39, 'sestiliardi'),
        # (1e36, 'sestilioni'),
        # (1e33, 'quintiliardi'),
        # (1e30, 'quintilioni'),
        # (1e27, 'quadriliardi'),
        # (1e24, 'quadrilioni'),    # yotta
        (1e21, 'triliardi'),  # zetta
        (1e18, 'trilioni'),  # exa
        (1e15, 'biliardi'),  # peta
        (1e12, 'bilioni'),  # tera
        (1e9, 'miliardi'),  # giga
        (1e6, 'milioni')  # mega
    ])

    multiplier = {}
    un_multiplier = {}

    for num in multipli_it:
        if num > 1000 and num <= 1e21:
            # plurali
            multiplier[multipli_it[num]] = int(num)
            # singolari - modificare per eccezioni *liardo
            if multipli_it[num][-5:-1] == 'iard':
                un_multiplier['un' + multipli_it[num][:-1] + 'o'] = int(num)
            else:
                un_multiplier['un' + multipli_it[num][:-1] + 'e'] = int(num)

    value = False

    # normalizza ordinali singoli o plurali -esimo -esimi
    if word[-5:-1] == 'esim':
        base = word[:-5]
        normalize_ita3 = {'tre': '', 'ttr': 'o', 'sei': '', 'ott': 'o'}
        normalize_ita2 = {'un': 'o', 'du': 'e', 'qu': 'e', 'tt': 'e',
                          'ov': 'e'}

        if base[-3:] in normalize_ita3:
            base += normalize_ita3[base[-3:]]
        elif base[-2:] in normalize_ita2:
            base += normalize_ita2[base[-2:]]

        word = base

    for item in un_multiplier:
        components = word.split(item, 1)
        if len(components) == 2:
            if not components[0]:  # inizia con un1^x
                if not components[1]:  # unmilione
                    word = str(int(un_multiplier[item]))
                else:  # unmilione + x
                    word = str(int(un_multiplier[item]) +
                               _extract_number_long_it(components[1]))

    for item in multiplier:
        components = word.split(item, 1)
        if len(components) == 2:
            if not components[0]:  # inizia con un1^x
                word = str(int(multiplier[item]) +
                           _extract_number_long_it(components[1]))
            else:
                if not components[1]:
                    word = str(_extract_number_long_it(components[0])) + '*' \
                           + str(int(multiplier[item]))
                else:
                    word = str(_extract_number_long_it(components[0])) + '*' \
                           + str(int(multiplier[item])) + '+' \
                           + str(_extract_number_long_it(components[1]))

    for item in tens:
        word = word.replace(item, '+' + str(tens[item]))

    for item in tens_short:
        word = word.replace(item, '+' + str(tens_short[item]))

    for item in nums_long:
        word = word.replace(item, '+' + str(nums_long[item]))

    word = word.replace('cento', '+1xx')
    word = word.replace('cent', '+1xx')
    word = word.replace('mille', '+1000')  # unmilionemille
    word = word.replace('mila', '*1000')  # unmilioneduemila

    for item in units:
        word = word.replace(item, '+' + str(units[item]))

    # normalizzo i cento
    occorrenze = word.count('+1xx')
    for _ in range(0, occorrenze):
        components = word.rsplit('+1xx', 1)
        if len(components[0]) > 1 and components[0].endswith('0'):
            word = components[0] + '+100' + components[1]
        else:
            word = components[0] + '*100' + components[1]

    components = word.rsplit('*1000', 1)
    if len(components) == 2:
        if components[0].startswith('*'):  # centomila
            components[0] = components[0][1:]
        word = str(_extract_number_long_it(components[0])) + \
               '*1000' + str(components[1])

    # gestione eccezioni
    if word.startswith('*') or word.startswith('+'):
        word = word[1:]

    addends = word.split('+')
    for c, _ in enumerate(addends):
        if '*' in addends[c]:
            factors = addends[c].split('*')
            result = int(factors[0]) * int(factors[1])
            if len(factors) == 3:
                result *= int(factors[2])
            addends[c] = str(result)

    # check if all token are numbers
    if all([s.isdecimal() for s in addends]):
        value = sum([int(s) for s in addends])
    else:
        value = False
    return value


def extract_number_it(text, short_scale=False, ordinals=False):
    """
    This function extracts a number from a text string,
    handles pronunciations in long scale and short scale

    https://en.wikipedia.org/wiki/Names_of_large_numbers

    Args:
        text (str): the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers, third=3 instead of 1/3
    Returns:
        (int) or (float) or False: The extracted number or False if no number
                                   was found

    """

    text = text.lower()
    string_num_ordinal_it = {}
    # first, second...
    if ordinals:
        if short_scale:
            for num in _SHORT_ORDINAL_STRING_IT:
                num_string = _SHORT_ORDINAL_STRING_IT[num]
                string_num_ordinal_it[num_string] = num
                _STRING_NUM_IT[num_string] = num
        else:
            for num in _LONG_ORDINAL_STRING_IT:
                num_string = _LONG_ORDINAL_STRING_IT[num]
                string_num_ordinal_it[num_string] = num
                _STRING_NUM_IT[num_string] = num

    # negate next number (-2 = 0 - 2)
    negatives = ['meno']  # 'negativo' non è usuale in italiano

    # multiply the previous number (one hundred = 1 * 100)
    multiplies = ['decina', 'decine', 'dozzina', 'dozzine',
                  'centinaia', 'centinaio', 'migliaia', 'migliaio', 'mila']

    # split sentence parse separately and sum ( 2 and a half = 2 + 0.5 )
    fraction_marker = [' e ']

    # decimal marker ( 1 point 5 = 1 + 0.5)
    decimal_marker = [' punto ', ' virgola ']

    if short_scale:
        for num in _SHORT_SCALE_IT:
            num_string = _SHORT_SCALE_IT[num]
            _STRING_NUM_IT[num_string] = num
            multiplies.append(num_string)
    else:
        for num in _LONG_SCALE_IT:
            num_string = _LONG_SCALE_IT[num]
            _STRING_NUM_IT[num_string] = num
            multiplies.append(num_string)

    # 2 e 3/4 ed altri casi
    for separator in fraction_marker:
        components = text.split(separator)
        zeros = 0

        if len(components) == 2:
            # count zeros in fraction part
            sub_components = components[1].split(' ')
            for element in sub_components:
                if element == 'zero' or element == '0':
                    zeros += 1
                else:
                    break
            # ensure first is not a fraction and second is a fraction
            num1 = extract_number_it(components[0])
            num2 = extract_number_it(components[1])
            if num1 is not None and num2 is not None \
                    and num1 >= 1 and 0 < num2 < 1:
                return num1 + num2
            # sette e quaranta  sette e zero zero due
            elif num1 is not None and num2 is not None \
                    and num1 >= 1 and num2 > 1:
                return num1 + num2 / pow(10, len(str(num2)) + zeros)

    # 2 punto 5
    for separator in decimal_marker:
        zeros = 0
        # count zeros in fraction part
        components = text.split(separator)

        if len(components) == 2:
            sub_components = components[1].split(' ')
            for element in sub_components:
                if element == 'zero' or element == '0':
                    zeros += 1
                else:
                    break

            number = int(extract_number_it(components[0]))
            decimal = int(extract_number_it(components[1]))
            if number is not None and decimal is not None:
                if '.' not in str(decimal):
                    return number + decimal / pow(10,
                                                  len(str(decimal)) + zeros)

    all_words = text.split()
    val = False
    prev_val = None
    to_sum = []
    for idx, word in enumerate(all_words):

        if not word:
            continue
        prev_word = all_words[idx - 1] if idx > 0 else ''
        next_word = all_words[idx + 1] if idx + 1 < len(all_words) else ''

        # is this word already a number ?
        if is_numeric(word):
            val = float(word)

        # is this word the name of a number ?
        if word in _STRING_NUM_IT:
            val = _STRING_NUM_IT[word]

        #  tre quarti  un quarto  trenta secondi
        if is_fractional_it(word) and prev_val:
            if word[:-1] == 'second' and not ordinals:
                val = prev_val * 2
            else:
                val = prev_val

        # is the prev word a number and should we multiply it?
        # twenty hundred, six hundred
        if word in multiplies:
            if not prev_val:
                prev_val = 1
            val = prev_val * val

        # is this a spoken fraction?
        # mezza tazza
        if val is False:
            val = is_fractional_it(word, short_scale=short_scale)

        # 2 quinti
        if not ordinals:
            next_value = is_fractional_it(next_word, short_scale=short_scale)
            if next_value:
                if not val:
                    val = 1
                val = val * next_value

        # is this a negative number?
        if val and prev_word and prev_word in negatives:
            val = 0 - val

        if not val:
            val = _extract_number_long_it(word)

        # let's make sure it isn't a fraction
        if not val:
            # look for fractions like '2/3'
            all_pieces = word.split('/')
            if look_for_fractions(all_pieces):
                val = float(all_pieces[0]) / float(all_pieces[1])
        else:
            prev_val = val
            # handle long numbers
            # six hundred sixty six
            # two million five hundred thousand
            if word in multiplies and next_word not in multiplies:
                to_sum.append(val)
                val = 0
                prev_val = 0
            elif _extract_number_long_it(word) > 100 and \
                    _extract_number_long_it(next_word) and \
                    next_word not in multiplies:
                to_sum.append(val)
                val = 0
                prev_val = 0

    if val is not None:
        for addend in to_sum:
            val = val + addend
    return val


def nice_number_it(number, speech=True, denominators=range(1, 21)):
    """ Italian helper for nice_number

    This function formats a float to human understandable functions. Like
    4.5 becomes "4 e un mezz" for speech and "4 1/2" for text

    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """

    result = convert_to_mixed_fraction(number, denominators)
    if not result:
        # Give up, just represent as a 3 decimal number
        return str(round(number, 3))

    whole, num, den = result

    if not speech:
        if num == 0:
            return str(whole)
        else:
            return '{} {}/{}'.format(whole, num, den)

    if num == 0:
        return str(whole)
    # denominatore
    den_str = _FRACTION_STRING_IT[den]
    # frazione
    if whole == 0:
        if num == 1:
            # un decimo
            return_string = 'un {}'.format(den_str)
        else:
            # tre mezzi
            return_string = '{} {}'.format(num, den_str)
    # interi  >10
    elif num == 1:
        # trenta e un
        return_string = '{} e un {}'.format(whole, den_str)
    # interi >10 con frazioni
    else:
        # venti e 3 decimi
        return_string = '{} e {} {}'.format(whole, num, den_str)

    # gestisce il plurale del denominatore
    if num > 1:
        return_string += 'i'
    else:
        return_string += 'o'

    return return_string


def pronounce_number_it(number, places=2, short_scale=False, scientific=False):
    """
    Convert a number to it's spoken equivalent
    adapted to italian fron en version

    For example, '5.2' would return 'cinque virgola due'

    Args:
        num(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
        short_scale (bool) : use short (True) or long scale (False)
            https://en.wikipedia.org/wiki/Names_of_large_numbers
        scientific (bool): pronounce in scientific notation
    Returns:
        (str): The pronounced number
    """
    num = number
    # gestione infinito
    if num == float("inf"):
        return "infinito"
    elif num == float("-inf"):
        return "meno infinito"

    if scientific:
        number = '%E' % num
        n, power = number.replace("+", "").split("E")
        power = int(power)
        if power != 0:
            return '{}{} per dieci elevato alla {}{}'.format(
                'meno ' if float(n) < 0 else '',
                pronounce_number_it(abs(float(n)), places, short_scale, False),
                'meno ' if power < 0 else '',
                pronounce_number_it(abs(power), places, short_scale, False))

    if short_scale:
        number_names = _NUM_STRING_IT.copy()
        number_names.update(_SHORT_SCALE_IT)
    else:
        number_names = _NUM_STRING_IT.copy()
        number_names.update(_LONG_SCALE_IT)

    digits = [number_names[n] for n in range(0, 20)]

    tens = [number_names[n] for n in range(10, 100, 10)]

    if short_scale:
        hundreds = [_SHORT_SCALE_IT[n] for n in _SHORT_SCALE_IT.keys()]
    else:
        hundreds = [_LONG_SCALE_IT[n] for n in _LONG_SCALE_IT.keys()]

    # deal with negatives
    result = ""
    if num < 0:
        result = "meno "
    num = abs(num)

    # check for a direct match
    if num in number_names:
        if num > 90:
            result += ""  # inizio stringa
        result += number_names[num]
    else:
        def _sub_thousand(n):
            assert 0 <= n <= 999
            if n <= 19:
                return digits[n]
            elif n <= 99:
                q, r = divmod(n, 10)
                _deci = tens[q - 1]
                _unit = r
                _partial = _deci
                if _unit > 0:
                    if _unit == 1 or _unit == 8:
                        _partial = _partial[:-1]  # ventuno  ventotto
                    _partial += number_names[_unit]
                return _partial
            else:
                q, r = divmod(n, 100)
                if q == 1:
                    _partial = "cento"
                else:
                    _partial = digits[q] + "cento"
                _partial += (
                    " " + _sub_thousand(r) if r else "")  # separa centinaia
                return _partial

        def _short_scale(n):
            if n >= max(_SHORT_SCALE_IT.keys()):
                return "numero davvero enorme"
            n = int(n)
            assert 0 <= n
            res = []
            for i, z in enumerate(_split_by(n, 1000)):
                if not z:
                    continue
                number = _sub_thousand(z)
                if i:
                    number += ""  # separa ordini grandezza
                    number += hundreds[i]
                res.append(number)

            return ", ".join(reversed(res))

        def _split_by(n, split=1000):
            assert 0 <= n
            res = []
            while n:
                n, r = divmod(n, split)
                res.append(r)
            return res

        def _long_scale(n):
            if n >= max(_LONG_SCALE_IT.keys()):
                return "numero davvero enorme"
            n = int(n)
            assert 0 <= n
            res = []
            for i, z in enumerate(_split_by(n, 1000000)):
                if not z:
                    continue
                number = pronounce_number_it(z, places, True, scientific)
                # strip off the comma after the thousand
                if i:
                    # plus one as we skip 'thousand'
                    # (and 'hundred', but this is excluded by index value)
                    number = number.replace(',', '')
                    number += " " + hundreds[i + 1]
                res.append(number)
            return ", ".join(reversed(res))

        if short_scale:
            result += _short_scale(num)
        else:
            result += _long_scale(num)

    # normalizza unità misura singole e 'ragionevoli' ed ad inizio stringa
    if result == 'mila':
        result = 'mille'
    if result == 'milioni':
        result = 'un milione'
    if result == 'miliardi':
        result = 'un miliardo'
    if result[0:7] == 'unomila':
        result = result.replace('unomila', 'mille', 1)
    if result[0:10] == 'unomilioni':
        result = result.replace('unomilioni', 'un milione', 1)
    # if result[0:11] == 'unomiliardi':
    # result = result.replace('unomiliardi', 'un miliardo', 1)

    # Deal with fractional part
    if not num == int(num) and places > 0:
        if abs(num) < 1.0 and (result == "meno " or not result):
            result += "zero"
        result += " virgola"
        _num_str = str(num)
        _num_str = _num_str.split(".")[1][0:places]
        for char in _num_str:
            result += " " + number_names[int(char)]
    return result



from collections import OrderedDict
from typing import List

from ovos_number_parser.util import (convert_to_mixed_fraction, look_for_fractions,
                                     is_numeric, tokenize, Token)

_ARTICLES_ES = {'el', 'la', 'los', 'las'}

_NUM_STRING_ES = {
    0: 'cero',
    1: 'uno',
    2: 'dos',
    3: 'tres',
    4: 'cuatro',
    5: 'cinco',
    6: 'seis',
    7: 'siete',
    8: 'ocho',
    9: 'nueve',
    10: 'diez',
    11: 'once',
    12: 'doce',
    13: 'trece',
    14: 'catorce',
    15: 'quince',
    16: 'dieciséis',
    17: 'diecisete',
    18: 'dieciocho',
    19: 'diecinueve',
    20: 'veinte',
    30: 'treinta',
    40: 'cuarenta',
    50: 'cincuenta',
    60: 'sesenta',
    70: 'setenta',
    80: 'ochenta',
    90: 'noventa'
}

_STRING_NUM_ES = {
    "cero": 0,
    "un": 1,
    "uno": 1,
    "una": 1,
    "dos": 2,
    "tres": 3,
    "trés": 3,
    "cuatro": 4,
    "cinco": 5,
    "seis": 6,
    "siete": 7,
    "ocho": 8,
    "nueve": 9,
    "diez": 10,
    "once": 11,
    "doce": 12,
    "trece": 13,
    "catorce": 14,
    "quince": 15,
    "dieciseis": 16,
    "dieciséis": 16,
    "diecisiete": 17,
    "dieciocho": 18,
    "diecinueve": 19,
    "veinte": 20,
    "veintiuno": 21,
    "veintidos": 22,
    "veintitres": 23,
    "veintidós": 22,
    "veintitrés": 23,
    "veinticuatro": 24,
    "veinticinco": 25,
    "veintiséis": 26,
    "veintiseis": 26,
    "veintisiete": 27,
    "veintiocho": 28,
    "veintinueve": 29,
    "treinta": 30,
    "cuarenta": 40,
    "cincuenta": 50,
    "sesenta": 60,
    "setenta": 70,
    "ochenta": 80,
    "noventa": 90,
    "cien": 100,
    "ciento": 100,
    "doscientos": 200,
    "doscientas": 200,
    "trescientos": 300,
    "trescientas": 300,
    "cuatrocientos": 400,
    "cuatrocientas": 400,
    "quinientos": 500,
    "quinientas": 500,
    "seiscientos": 600,
    "seiscientas": 600,
    "setecientos": 700,
    "setecientas": 700,
    "ochocientos": 800,
    "ochocientas": 800,
    "novecientos": 900,
    "novecientas": 900,
    "mil": 1000}

_FRACTION_STRING_ES = {
    2: 'medio',
    3: 'tercio',
    4: 'cuarto',
    5: 'quinto',
    6: 'sexto',
    7: 'séptimo',
    8: 'octavo',
    9: 'noveno',
    10: 'décimo',
    11: 'onceavo',
    12: 'doceavo',
    13: 'treceavo',
    14: 'catorceavo',
    15: 'quinceavo',
    16: 'dieciseisavo',
    17: 'diecisieteavo',
    18: 'dieciochoavo',
    19: 'diecinueveavo',
    20: 'veinteavo'
}

# https://www.grobauer.at/es_eur/zahlnamen.php
_LONG_SCALE_ES = OrderedDict([
    (100, 'centena'),
    (1000, 'millar'),
    (1000000, 'millón'),
    (1e9, "millardo"),
    (1e12, "billón"),
    (1e18, 'trillón'),
    (1e24, "cuatrillón"),
    (1e30, "quintillón"),
    (1e36, "sextillón"),
    (1e42, "septillón"),
    (1e48, "octillón"),
    (1e54, "nonillón"),
    (1e60, "decillón"),
    (1e66, "undecillón"),
    (1e72, "duodecillón"),
    (1e78, "tredecillón"),
    (1e84, "cuatrodecillón"),
    (1e90, "quindecillón"),
    (1e96, "sexdecillón"),
    (1e102, "septendecillón"),
    (1e108, "octodecillón"),
    (1e114, "novendecillón"),
    (1e120, "vigintillón"),
    (1e306, "unquinquagintillón"),
    (1e312, "duoquinquagintillón"),
    (1e336, "sexquinquagintillón"),
    (1e366, "unsexagintillón")
])

_SHORT_SCALE_ES = OrderedDict([
    (100, 'centena'),
    (1000, 'millar'),
    (1000000, 'millón'),
    (1e9, "billón"),
    (1e12, 'trillón'),
    (1e15, "cuatrillón"),
    (1e18, "quintillón"),
    (1e21, "sextillón"),
    (1e24, "septillón"),
    (1e27, "octillón"),
    (1e30, "nonillón"),
    (1e33, "decillón"),
    (1e36, "undecillón"),
    (1e39, "duodecillón"),
    (1e42, "tredecillón"),
    (1e45, "cuatrodecillón"),
    (1e48, "quindecillón"),
    (1e51, "sexdecillón"),
    (1e54, "septendecillón"),
    (1e57, "octodecillón"),
    (1e60, "novendecillón"),
    (1e63, "vigintillón"),
    (1e66, "unvigintillón"),
    (1e69, "uuovigintillón"),
    (1e72, "tresvigintillón"),
    (1e75, "quattuorvigintillón"),
    (1e78, "quinquavigintillón"),
    (1e81, "qesvigintillón"),
    (1e84, "septemvigintillón"),
    (1e87, "octovigintillón"),
    (1e90, "novemvigintillón"),
    (1e93, "trigintillón"),
    (1e96, "untrigintillón"),
    (1e99, "duotrigintillón"),
    (1e102, "trestrigintillón"),
    (1e105, "quattuortrigintillón"),
    (1e108, "quinquatrigintillón"),
    (1e111, "sestrigintillón"),
    (1e114, "septentrigintillón"),
    (1e117, "octotrigintillón"),
    (1e120, "noventrigintillón"),
    (1e123, "quadragintillón"),
    (1e153, "quinquagintillón"),
    (1e183, "sexagintillón"),
    (1e213, "septuagintillón"),
    (1e243, "octogintillón"),
    (1e273, "nonagintillón"),
    (1e303, "centillón"),
    (1e306, "uncentillón"),
    (1e309, "duocentillón"),
    (1e312, "trescentillón"),
    (1e333, "decicentillón"),
    (1e336, "undecicentillón"),
    (1e363, "viginticentillón"),
    (1e366, "unviginticentillón"),
    (1e393, "trigintacentillón"),
    (1e423, "quadragintacentillón"),
    (1e453, "quinquagintacentillón"),
    (1e483, "sexagintacentillón"),
    (1e513, "septuagintacentillón"),
    (1e543, "ctogintacentillón"),
    (1e573, "nonagintacentillón"),
    (1e603, "ducentillón"),
    (1e903, "trecentillón"),
    (1e1203, "quadringentillón"),
    (1e1503, "quingentillón"),
    (1e1803, "sexcentillón"),
    (1e2103, "septingentillón"),
    (1e2403, "octingentillón"),
    (1e2703, "nongentillón"),
    (1e3003, "millinillón")
])

# TODO: female forms.
_ORDINAL_STRING_BASE_ES = {
    1: 'primero',
    2: 'segundo',
    3: 'tercero',
    4: 'cuarto',
    5: 'quinto',
    6: 'sexto',
    7: 'séptimo',
    8: 'octavo',
    9: 'noveno',
    10: 'décimo',
    11: 'undécimo',
    12: 'duodécimo',
    13: 'decimotercero',
    14: 'decimocuarto',
    15: 'decimoquinto',
    16: 'decimosexto',
    17: 'decimoséptimo',
    18: 'decimoctavo',
    19: 'decimonoveno',
    20: 'vigésimo',
    30: 'trigésimo',
    40: "cuadragésimo",
    50: "quincuagésimo",
    60: "sexagésimo",
    70: "septuagésimo",
    80: "octogésimo",
    90: "nonagésimo",
    10e3: "centésimó",
    1e3: "milésimo"
}

_SHORT_ORDINAL_STRING_ES = {
    1e6: "millonésimo",
    1e9: "milmillonésimo",
    1e12: "billonésimo",
    1e15: "milbillonésimo",
    1e18: "trillonésimo",
    1e21: "miltrillonésimo",
    1e24: "cuatrillonésimo",
    1e27: "milcuatrillonésimo",
    1e30: "quintillonésimo",
    1e33: "milquintillonésimo"
    # TODO > 1e-33
}
_SHORT_ORDINAL_STRING_ES.update(_ORDINAL_STRING_BASE_ES)

_LONG_ORDINAL_STRING_ES = {
    1e6: "millonésimo",
    1e12: "billionth",
    1e18: "trillonésimo",
    1e24: "cuatrillonésimo",
    1e30: "quintillonésimo",
    1e36: "sextillonésimo",
    1e42: "septillonésimo",
    1e48: "octillonésimo",
    1e54: "nonillonésimo",
    1e60: "decillonésimo"
    # TODO > 1e60
}
_LONG_ORDINAL_STRING_ES.update(_ORDINAL_STRING_BASE_ES)


def is_fractional_es(input_str, short_scale=True):
    """
    This function takes the given text and checks if it is a fraction.

    Args:
        text (str): the string to check if fractional

        short_scale (bool): use short scale if True, long scale if False
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction

    """
    if input_str.endswith('s', -1):
        input_str = input_str[:len(input_str) - 1]  # e.g. "fifths"

    aFrac = {"medio": 2, "media": 2, "tercio": 3, "cuarto": 4,
             "cuarta": 4, "quinto": 5, "quinta": 5, "sexto": 6, "sexta": 6,
             "séptimo": 7, "séptima": 7, "octavo": 8, "octava": 8,
             "noveno": 9, "novena": 9, "décimo": 10, "décima": 10,
             "onceavo": 11, "onceava": 11, "doceavo": 12, "doceava": 12}

    if input_str.lower() in aFrac:
        return 1.0 / aFrac[input_str]
    if (input_str == "vigésimo" or input_str == "vigésima"):
        return 1.0 / 20
    if (input_str == "trigésimo" or input_str == "trigésima"):
        return 1.0 / 30
    if (input_str == "centésimo" or input_str == "centésima"):
        return 1.0 / 100
    if (input_str == "milésimo" or input_str == "milésima"):
        return 1.0 / 1000
    return False


def extract_number_es(text, short_scale=True, ordinals=False):
    """
    This function prepares the given text for parsing by making
    numbers consistent, getting rid of contractions, etc.
    Args:
        text (str): the string to normalize
    Returns:
        (int) or (float): The value of extracted number

    """
    # TODO: short_scale and ordinals don't do anything here.
    # The parameters are present in the function signature for API compatibility
    # reasons.
    #
    # Returns incorrect output on certain fractional phrases like, "cuarto de dos"
    #  TODO: numbers greater than 999999
    aWords = text.lower().split()
    count = 0
    result = None
    while count < len(aWords):
        val = 0
        word = aWords[count]
        next_next_word = None
        if count + 1 < len(aWords):
            next_word = aWords[count + 1]
            if count + 2 < len(aWords):
                next_next_word = aWords[count + 2]
        else:
            next_word = None

        # is current word a number?
        if word in _STRING_NUM_ES:
            val = _STRING_NUM_ES[word]
        elif word.isdigit():  # doesn't work with decimals
            val = int(word)
        elif is_numeric(word):
            val = float(word)
        elif is_fractional_es(word):
            if not result:
                result = 1
            result = result * is_fractional_es(word)
            count += 1
            continue

        if not val:
            # look for fractions like "2/3"
            aPieces = word.split('/')
            # if (len(aPieces) == 2 and is_numeric(aPieces[0])
            #   and is_numeric(aPieces[1])):
            if look_for_fractions(aPieces):
                val = float(aPieces[0]) / float(aPieces[1])

        if val:
            if result is None:
                result = 0
            # handle fractions
            if next_word != "avos":
                result = val
            else:
                result = float(result) / float(val)

        if next_word is None:
            break

        # number word and fraction
        ands = ["y"]
        if next_word in ands:
            zeros = 0
            if result is None:
                count += 1
                continue
            newWords = aWords[count + 2:]
            newText = ""
            for word in newWords:
                newText += word + " "

            afterAndVal = extract_number_es(newText[:-1])
            if afterAndVal:
                if result < afterAndVal or result < 20:
                    while afterAndVal > 1:
                        afterAndVal = afterAndVal / 10.0
                    for word in newWords:
                        if word == "cero" or word == "0":
                            zeros += 1
                        else:
                            break
                for _ in range(0, zeros):
                    afterAndVal = afterAndVal / 10.0
                result += afterAndVal
                break
        elif next_next_word is not None:
            if next_next_word in ands:
                newWords = aWords[count + 3:]
                newText = ""
                for word in newWords:
                    newText += word + " "
                afterAndVal = extract_number_es(newText[:-1])
                if afterAndVal:
                    if result is None:
                        result = 0
                    result += afterAndVal
                    break

        decimals = ["punto", "coma", ".", ","]
        if next_word in decimals:
            zeros = 0
            newWords = aWords[count + 2:]
            newText = ""
            for word in newWords:
                newText += word + " "
            for word in newWords:
                if word == "cero" or word == "0":
                    zeros += 1
                else:
                    break
            afterDotVal = str(extract_number_es(newText[:-1]))
            afterDotVal = zeros * "0" + afterDotVal
            result = float(str(result) + "." + afterDotVal)
            break
        count += 1

    # Return the $str with the number related words removed
    # (now empty strings, so strlen == 0)
    # aWords = [word for word in aWords if len(word) > 0]
    # text = ' '.join(aWords)
    if "." in str(result):
        integer, dec = str(result).split(".")
        # cast float to int
        if dec == "0":
            result = int(integer)

    return result or False


def _es_number_parse(words, i):
    # TODO Not parsing 'cero'

    def es_cte(i, s):
        if i < len(words) and s == words[i]:
            return s, i + 1
        return None

    def es_number_word(i, mi, ma):
        if i < len(words):
            v = _STRING_NUM_ES.get(words[i])
            if v and v >= mi and v <= ma:
                return v, i + 1
        return None

    def es_number_1_99(i):
        r1 = es_number_word(i, 1, 29)
        if r1:
            return r1

        r1 = es_number_word(i, 30, 90)
        if r1:
            v1, i1 = r1
            r2 = es_cte(i1, "y")
            if r2:
                i2 = r2[1]
                r3 = es_number_word(i2, 1, 9)
                if r3:
                    v3, i3 = r3
                    return v1 + v3, i3
            return r1
        return None

    def es_number_1_999(i):
        # [2-9]cientos [1-99]?
        r1 = es_number_word(i, 100, 900)
        if r1:
            v1, i1 = r1
            r2 = es_number_1_99(i1)
            if r2:
                v2, i2 = r2
                return v1 + v2, i2
            else:
                return r1

        # [1-99]
        r1 = es_number_1_99(i)
        if r1:
            return r1

        return None

    def es_number(i):
        # check for cero
        r1 = es_number_word(i, 0, 0)
        if r1:
            return r1

        # check for [1-999] (mil [0-999])?
        r1 = es_number_1_999(i)
        if r1:
            v1, i1 = r1
            r2 = es_cte(i1, "mil")
            if r2:
                i2 = r2[1]
                r3 = es_number_1_999(i2)
                if r3:
                    v3, i3 = r3
                    return v1 * 1000 + v3, i3
                else:
                    return v1 * 1000, i2
            else:
                return r1
        return None

    return es_number(i)


def nice_number_es(number, speech=True, denominators=range(1, 21)):
    """ Spanish helper for nice_number

    This function formats a float to human understandable functions. Like
    4.5 becomes "4 y medio" for speech and "4 1/2" for text

    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """
    strNumber = ""
    whole = 0
    num = 0
    den = 0

    result = convert_to_mixed_fraction(number, denominators)

    if not result:
        # Give up, just represent as a 3 decimal number
        whole = round(number, 3)
    else:
        whole, num, den = result

    if not speech:
        if num == 0:
            strNumber = '{:,}'.format(whole)
            strNumber = strNumber.replace(",", " ")
            strNumber = strNumber.replace(".", ",")
            return strNumber
        else:
            return '{} {}/{}'.format(whole, num, den)
    else:
        if num == 0:
            # if the number is not a fraction, nothing to do
            strNumber = str(whole)
            strNumber = strNumber.replace(".", ",")
            return strNumber
        den_str = _FRACTION_STRING_ES[den]
        # if it is not an integer
        if whole == 0:
            # if there is no whole number
            if num == 1:
                # if numerator is 1, return "un medio", for example
                strNumber = 'un {}'.format(den_str)
            else:
                # else return "cuatro tercios", for example
                strNumber = '{} {}'.format(num, den_str)
        elif num == 1:
            # if there is a whole number and numerator is 1
            if den == 2:
                # if denominator is 2, return "1 y medio", for example
                strNumber = '{} y {}'.format(whole, den_str)
            else:
                # else return "1 y 1 tercio", for example
                strNumber = '{} y 1 {}'.format(whole, den_str)
        else:
            # else return "2 y 3 cuarto", for example
            strNumber = '{} y {} {}'.format(whole, num, den_str)
        if num > 1 and den != 3:
            # if the numerator is greater than 1 and the denominator
            # is not 3 ("tercio"), add an s for plural
            strNumber += 's'

    return strNumber


def pronounce_number_es(number, places=2):
    """
    Convert a number to it's spoken equivalent

    For example, '5.2' would return 'cinco coma dos'

    Args:
        num(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
    Returns:
        (str): The pronounced number
    """
    if abs(number) >= 100:
        # TODO: Soporta a números por encima de 100
        return str(number)

    result = ""
    if number < 0:
        result = "menos "
    number = abs(number)

    # del 21 al 29 tienen una pronunciación especial
    if 20 <= number <= 29:
        tens = int(number - int(number) % 10)
        ones = int(number - tens)
        result += _NUM_STRING_ES[tens]
        if ones > 0:
            result = result[:-1]
            # a veinte le quitamos la "e" final para construir los
            # números del 21 - 29. Pero primero tenemos en cuenta
            # las excepciones: 22, 23 y 26, que llevan tilde.
            if ones == 2:
                result += "idós"
            elif ones == 3:
                result += "itrés"
            elif ones == 6:
                result += "iséis"
            else:
                result += "i" + _NUM_STRING_ES[ones]
    elif number >= 30:  # de 30 en adelante
        tens = int(number - int(number) % 10)
        ones = int(number - tens)
        result += _NUM_STRING_ES[tens]
        if ones > 0:
            result += " y " + _NUM_STRING_ES[ones]
    else:
        result += _NUM_STRING_ES[int(number)]

    # Deal with decimal part, in spanish is commonly used the comma
    # instead the dot. Decimal part can be written both with comma
    # and dot, but when pronounced, its pronounced "coma"
    if not number == int(number) and places > 0:
        if abs(number) < 1.0 and (result == "menos " or not result):
            result += "cero"
        result += " coma"
        _num_str = str(number)
        _num_str = _num_str.split(".")[1][0:places]
        for char in _num_str:
            result += " " + _NUM_STRING_ES[int(char)]
    return result


def numbers_to_digits_es(utterance: str) -> str:
    """
    Replace written numbers in a Spanish text with their digit equivalents.

    Args:
        utterance (str): Input string possibly containing written numbers.

    Returns:
        str: Text with written numbers replaced by digits.
    """
    # TODO - above twenty it's ambiguous, "twenty one" is 2 words but only 1 number
    number_replacements = {
        "uno": "1", "dos": "2", "tres": "3", "cuatro": "4",
        "cinco": "5", "seis": "6", "siete": "7", "ocho": "8", "nueve": "9",
        "diez": "10", "once": "11", "doce": "12", "trece": "13", "catorce": "14",
        "quince": "15", "dieciséis": "16", "diecisiete": "17", "dieciocho": "18",
        "diecinueve": "19", "veinte": "20"
        # Extend this dictionary for higher numbers as needed
    }
    words: List[Token] = tokenize(utterance)
    for idx, tok in enumerate(words):
        if tok.word in number_replacements:
            words[idx] = number_replacements[tok.word]
        else:
            words[idx] = tok.word
    return " ".join(words)



from typing import List

from ovos_number_parser.util import (convert_to_mixed_fraction, look_for_fractions,
                                     is_numeric, tokenize, Token)

_NUMBERS_CA = {
    "zero": 0,
    "u": 1,
    "un": 1,
    "una": 1,
    "uns": 1,
    "unes": 1,
    "primer": 1,
    "primera": 1,
    "segon": 2,
    "segona": 2,
    "tercer": 3,
    "tercera": 3,
    "dos": 2,
    "dues": 2,
    "tres": 3,
    "quatre": 4,
    "cinc": 5,
    "sis": 6,
    "set": 7,
    "vuit": 8,
    "huit": 8,
    "nou": 9,
    "deu": 10,
    "onze": 11,
    "dotze": 12,
    "tretze": 13,
    "catorze": 14,
    "quinze": 15,
    "setze": 16,
    "disset": 17,
    "divuit": 18,
    "dinou": 19,
    "vint": 20,
    "trenta": 30,
    "quaranta": 40,
    "cinquanta": 50,
    "seixanta": 60,
    "setanta": 70,
    "vuitanta": 80,
    "noranta": 90,
    "cent": 100,
    "cents": 100,
    "dos-cents": 200,
    "dues-centes": 200,
    "tres-cents": 300,
    "tres-centes": 300,
    "quatre-cents": 400,
    "quatre-centes": 400,
    "cinc-cents": 500,
    "cinc-centes": 500,
    "sis-cents": 600,
    "sis-centes": 600,
    "set--cents": 700,
    "set-centes": 700,
    "vuit-cents": 800,
    "vuit-centes": 800,
    "nou-cents": 900,
    "nou-centes": 900,
    "mil": 1000,
    "milió": 1000000
}

_FRACTION_STRING_CA = {
    2: 'mig',
    3: 'terç',
    4: 'quart',
    5: 'cinquè',
    6: 'sisè',
    7: 'setè',
    8: 'vuitè',
    9: 'novè',
    10: 'desè',
    11: 'onzè',
    12: 'dotzè',
    13: 'tretzè',
    14: 'catorzè',
    15: 'quinzè',
    16: 'setzè',
    17: 'dissetè',
    18: 'divuitè',
    19: 'dinovè',
    20: 'vintè',
    30: 'trentè',
    100: 'centè',
    1000: 'milè'
}

_NUM_STRING_CA = {
    0: 'zero',
    1: 'un',
    2: 'dos',
    3: 'tres',
    4: 'quatre',
    5: 'cinc',
    6: 'sis',
    7: 'set',
    8: 'vuit',
    9: 'nou',
    10: 'deu',
    11: 'onze',
    12: 'dotze',
    13: 'tretze',
    14: 'catorze',
    15: 'quinze',
    16: 'setze',
    17: 'disset',
    18: 'divuit',
    19: 'dinou',
    20: 'vint',
    30: 'trenta',
    40: 'quaranta',
    50: 'cinquanta',
    60: 'seixanta',
    70: 'setanta',
    80: 'vuitanta',
    90: 'noranta'
}

_TENS_CA = {
    "vint": 20,
    "trenta": 30,
    "quaranta": 40,
    "cinquanta": 50,
    "seixanta": 60,
    "setanta": 70,
    "vuitanta": 80,
    "huitanta": 80,
    "noranta": 90
}

_AFTER_TENS_CA = {
    "u": 1,
    "un": 1,
    "dos": 2,
    "dues": 2,
    "tres": 3,
    "quatre": 4,
    "cinc": 5,
    "sis": 6,
    "set": 7,
    "vuit": 8,
    "huit": 8,
    "nou": 9
}

_BEFORE_HUNDREDS_CA = {
    "dos": 2,
    "dues": 2,
    "tres": 3,
    "quatre": 4,
    "cinc": 5,
    "sis": 6,
    "set": 7,
    "vuit": 8,
    "huit": 8,
    "nou": 9,
}

_HUNDREDS_CA = {
    "cent": 100,
    "cents": 100,
    "centes": 100
}


def nice_number_ca(number, speech, denominators=range(1, 21)):
    """ Catalan helper for nice_number

    This function formats a float to human understandable functions. Like
    4.5 becomes "4 i mig" for speech and "4 1/2" for text

    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """

    result = convert_to_mixed_fraction(number, denominators)
    if not result:
        # Give up, just represent as a 3 decimal number
        return str(round(number, 3))

    whole, num, den = result

    if not speech:
        if num == 0:
            # TODO: Number grouping?  E.g. "1,000,000"
            return str(whole)
        else:
            return '{} {}/{}'.format(whole, num, den)

    if num == 0:
        return str(whole)
    # denominador
    den_str = _FRACTION_STRING_CA[den]
    # fraccions
    if whole == 0:
        if num == 1:
            # un desè
            return_string = 'un {}'.format(den_str)
        else:
            # tres mig
            return_string = '{} {}'.format(num, den_str)
    # inteiros >10
    elif num == 1:
        # trenta-un
        return_string = '{}-{}'.format(whole, den_str)
    # inteiros >10 com fracções
    else:
        # vint i 3 desens
        return_string = '{} i {} {}'.format(whole, num, den_str)
    # plural
    if num > 1:
        return_string += 's'
    return return_string


def pronounce_number_ca(number, places=2):
    """
    Convert a number to it's spoken equivalent
     For example, '5.2' would return 'cinc coma dos'
     Args:
        number(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
    Returns:
        (str): The pronounced number
    """
    if abs(number) >= 100:
        # TODO: Support n > 100
        return str(number)

    result = ""
    if number < 0:
        result = "menys "
    number = abs(number)

    if number >= 20:
        tens = int(number - int(number) % 10)
        ones = int(number - tens)
        result += _NUM_STRING_CA[tens]
        if ones > 0:
            if tens == 20:
                result += "-i-" + _NUM_STRING_CA[ones]
            else:
                result += "-" + _NUM_STRING_CA[ones]
    else:
        result += _NUM_STRING_CA[int(number)]

    # Deal with decimal part, in Catalan is commonly used the comma
    # instead the dot. Decimal part can be written both with comma
    # and dot, but when pronounced, its pronounced "coma"
    if not number == int(number) and places > 0:
        if abs(number) < 1.0 and (result == "menys " or not result):
            result += "zero"
        result += " coma"
        _num_str = str(number)
        _num_str = _num_str.split(".")[1][0:places]
        for char in _num_str:
            result += " " + _NUM_STRING_CA[int(char)]
    return result


def is_fractional_ca(input_str, short_scale=True):
    """
    This function takes the given text and checks if it is a fraction.

    Args:
        input_str (str): the string to check if fractional
        short_scale (bool): use short scale if True, long scale if False
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction

    """
    if input_str.endswith('é', -1):
        input_str = input_str[:len(input_str) - 1] + "è"  # e.g. "cinqué -> cinquè"
    elif input_str.endswith('ena', -3):
        input_str = input_str[:len(input_str) - 3] + "è"  # e.g. "cinquena -> cinquè"
    elif input_str.endswith('ens', -3):
        input_str = input_str[:len(input_str) - 3] + "è"  # e.g. "cinquens -> cinquè"
    elif input_str.endswith('enes', -4):
        input_str = input_str[:len(input_str) - 4] + "è"  # e.g. "cinquenes -> cinquè"
    elif input_str.endswith('os', -2):
        input_str = input_str[:len(input_str) - 2]  # e.g. "terços -> terç"
    elif (input_str == 'terceres' or input_str == 'tercera'):
        input_str = "terç"  # e.g. "tercer -> terç"
    elif (input_str == 'mitges' or input_str == 'mitja'):
        input_str = "mig"  # e.g. "mitges -> mig"
    elif (input_str == 'meitat' or input_str == 'meitats'):
        input_str = "mig"  # e.g. "mitges -> mig"
    elif input_str.endswith('a', -1):
        input_str = input_str[:len(input_str) - 1]  # e.g. "quarta -> quart"
    elif input_str.endswith('es', -2):
        input_str = input_str[:len(input_str) - 2]  # e.g. "quartes -> quartes"
    elif input_str.endswith('s', -1):
        input_str = input_str[:len(input_str) - 1]  # e.g. "quarts -> quart"

    aFrac = ["mig", "terç", "quart", "cinquè", "sisè", "sètè", "vuitè", "novè",
             "desè", "onzè", "dotzè", "tretzè", "catorzè", "quinzè", "setzè",
             "dissetè", "divuitè", "dinovè"]

    if input_str.lower() in aFrac:
        return 1.0 / (aFrac.index(input_str) + 2)
    if input_str == "vintè":
        return 1.0 / 20
    if input_str == "trentè":
        return 1.0 / 30
    if input_str == "centè":
        return 1.0 / 100
    if input_str == "milè":
        return 1.0 / 1000
    if (input_str == "vuitè" or input_str == "huitè"):
        return 1.0 / 8
    if (input_str == "divuitè" or input_str == "dihuitè"):
        return 1.0 / 18

    return False


def extract_number_ca(text, short_scale=True, ordinals=False):
    """
    This function prepares the given text for parsing by making
    numbers consistent, getting rid of contractions, etc.
    Args:
        text (str): the string to normalize
    Returns:
        (int) or (float): The value of extracted number

    """
    # TODO: short_scale and ordinals don't do anything here.
    # The parameters are present in the function signature for API compatibility
    # reasons.
    text = text.lower()
    aWords = text.split()
    count = 0
    result = None
    while count < len(aWords):
        val = 0
        word = aWords[count]
        next_next_word = None
        if count + 1 < len(aWords):
            next_word = aWords[count + 1]
            if count + 2 < len(aWords):
                next_next_word = aWords[count + 2]
        else:
            next_word = None

        # is current word a number?
        if word in _NUMBERS_CA:
            val = _NUMBERS_CA[word]
        elif '-' in word:
            wordparts = word.split('-')
            # trenta-cinc > 35
            if len(wordparts) == 2 and (wordparts[0] in _TENS_CA and wordparts[1] in _AFTER_TENS_CA):
                val = _TENS_CA[wordparts[0]] + _AFTER_TENS_CA[wordparts[1]]
            # vint-i-dues > 22
            elif len(wordparts) == 3 and wordparts[1] == 'i' and (
                    wordparts[0] in _TENS_CA and wordparts[2] in _AFTER_TENS_CA):
                val = _TENS_CA[wordparts[0]] + _AFTER_TENS_CA[wordparts[2]]
            # quatre-centes > 400
            elif len(wordparts) == 2 and (wordparts[0] in _BEFORE_HUNDREDS_CA and wordparts[1] in _HUNDREDS_CA):
                val = _BEFORE_HUNDREDS_CA[wordparts[0]] * 100

        elif word.isdigit():  # doesn't work with decimals
            val = int(word)
        elif is_numeric(word):
            val = float(word)
        elif is_fractional_ca(word):
            if not result:
                result = 1
            result = result * is_fractional_ca(word)
            count += 1
            continue

        if not val:
            # look for fractions like "2/3"
            aPieces = word.split('/')
            # if (len(aPieces) == 2 and is_numeric(aPieces[0])
            #   and is_numeric(aPieces[1])):
            if look_for_fractions(aPieces):
                val = float(aPieces[0]) / float(aPieces[1])

        if val:
            if result is None:
                result = 0
            # handle fractions
            # TODO: caution, review use of "ens" word
            if next_word != "ens":
                result += val
            else:
                result = float(result) / float(val)

        if next_word is None:
            break

        # number word and fraction
        ands = ["i"]
        if next_word in ands:
            zeros = 0
            if result is None:
                count += 1
                continue
            newWords = aWords[count + 2:]
            newText = ""
            for word in newWords:
                newText += word + " "

            afterAndVal = extract_number_ca(newText[:-1])
            if afterAndVal:
                if result < afterAndVal or result < 20:
                    while afterAndVal > 1:
                        afterAndVal = afterAndVal / 10.0
                    for word in newWords:
                        if word == "zero" or word == "0":
                            zeros += 1
                        else:
                            break
                for _ in range(0, zeros):
                    afterAndVal = afterAndVal / 10.0
                result += afterAndVal
                break
        elif next_next_word is not None:
            if next_next_word in ands:
                newWords = aWords[count + 3:]
                newText = ""
                for word in newWords:
                    newText += word + " "
                afterAndVal = extract_number_ca(newText[:-1])
                if afterAndVal:
                    if result is None:
                        result = 0
                    result += afterAndVal
                    break

        decimals = ["coma", "amb", "punt", ".", ","]
        if next_word in decimals:
            zeros = 0
            newWords = aWords[count + 2:]
            newText = ""
            for word in newWords:
                newText += word + " "
            for word in newWords:
                if word == "zero" or word == "0":
                    zeros += 1
                else:
                    break
            afterDotVal = str(extract_number_ca(newText[:-1]))
            afterDotVal = zeros * "0" + afterDotVal
            result = float(str(result) + "." + afterDotVal)
            break
        count += 1

    # Return the $str with the number related words removed
    # (now empty strings, so strlen == 0)
    # aWords = [word for word in aWords if len(word) > 0]
    # text = ' '.join(aWords)
    if "." in str(result):
        integer, dec = str(result).split(".")
        # cast float to int
        if dec == "0":
            result = int(integer)

    return result or False


def numbers_to_digits_ca(utterance: str) -> str:
    """
    Substitueix els números escrits en un text en català per les seves equivalents en xifres.

    Args:
        utterance (str): Cadena d'entrada que possiblement conté números escrits.

    Returns:
        str: Text amb els números escrits substituïts per xifres.
    """
    # TODO - above twenty it's ambiguous, "twenty one" is 2 words but only 1 number
    number_replacements = {
        "un": "1", "dos": "2", "tres": "3", "quatre": "4",
        "cinc": "5", "sis": "6", "set": "7", "vuit": "8", "nou": "9",
        "deu": "10", "onze": "11", "dotze": "12", "tretze": "13", "catorze": "14",
        "quinze": "15", "setze": "16", "disset": "17", "divuit": "18",
        "dinou": "19", "vint": "20"
        # Amplieu aquest diccionari per a números més alts si és necessari
    }
    words: List[Token] = tokenize(utterance)
    for idx, tok in enumerate(words):
        if tok.word in number_replacements:
            words[idx] = number_replacements[tok.word]
        else:
            words[idx] = tok.word
    return " ".join(words)



import re
from enum import Enum
from typing import List, Union, Dict, Tuple

from ovos_number_parser.util import Scale, GrammaticalGender, DigitPronunciation


class PortugueseVariant(str, Enum):
    """
    Defines the Portuguese variant for spelling.
    - BR: Brazilian Portuguese (e.g., bilhão, dezesseis).
    - PT: European Portuguese (e.g., bilião, dezasseis).
    """
    BR = "br"
    PT = "pt"


# --- Base Pronunciation Dictionaries (Variant-aware) ---
# Dictionaries for units, tens, and hundreds, separated by Portuguese variant.

_UNITS: Dict[int, str] = {
    1: 'um', 2: 'dois', 3: 'três', 4: 'quatro', 5: 'cinco', 6: 'seis',
    7: 'sete', 8: 'oito', 9: 'nove'
}

_TENS_BR: Dict[int, str] = {
    10: 'dez', 11: 'onze', 12: 'doze', 13: 'treze', 14: 'catorze',
    15: 'quinze', 16: 'dezesseis', 17: 'dezessete', 18: 'dezoito',
    19: 'dezenove', 20: 'vinte', 30: 'trinta', 40: 'quarenta',
    50: 'cinquenta', 60: 'sessenta', 70: 'setenta', 80: 'oitenta',
    90: 'noventa'
}

_TENS_PT: Dict[int, str] = {
    10: 'dez', 11: 'onze', 12: 'doze', 13: 'treze', 14: 'catorze',
    15: 'quinze', 16: 'dezasseis', 17: 'dezassete', 18: 'dezoito',
    19: 'dezanove', 20: 'vinte', 30: 'trinta', 40: 'quarenta',
    50: 'cinquenta', 60: 'sessenta', 70: 'setenta', 80: 'oitenta',
    90: 'noventa'
}

_HUNDREDS: Dict[int, str] = {
    100: 'cem', 200: 'duzentos', 300: 'trezentos', 400: 'quatrocentos',
    500: 'quinhentos', 600: 'seiscentos', 700: 'setecentos',
    800: 'oitocentos', 900: 'novecentos'
}

_FRACTION_STRING_PT: Dict[int, str] = {
    2: 'meio', 3: 'terço', 4: 'quarto', 5: 'quinto', 6: 'sexto',
    7: 'sétimo', 8: 'oitavo', 9: 'nono', 10: 'décimo',
    11: 'onze avos', 12: 'doze avos', 13: 'treze avos', 14: 'catorze avos',
    15: 'quinze avos', 16: 'dezasseis avos', 17: 'dezassete avos',
    18: 'dezoito avos', 19: 'dezanove avos',
    20: 'vigésimo', 30: 'trigésimo', 100: 'centésimo', 1000: 'milésimo'
}

_FEMALE_NUMS = {
    "uma": 1,
    "duas": 2
}

# --- Ordinal Pronunciation Dictionaries (Masculine Base) ---
# These dictionaries are for masculine forms. The feminine form
# is generated by replacing the final 'o' with 'a'.
_ORDINAL_UNITS_MASC: Dict[int, str] = {
    1: 'primeiro', 2: 'segundo', 3: 'terceiro', 4: 'quarto', 5: 'quinto',
    6: 'sexto', 7: 'sétimo', 8: 'oitavo', 9: 'nono'
}

_ORDINAL_TENS_MASC: Dict[int, str] = {
    10: 'décimo', 20: 'vigésimo', 30: 'trigésimo', 40: 'quadragésimo',
    50: 'quinquagésimo', 60: 'sexagésimo', 70: 'septuagésimo',
    80: 'octogésimo', 90: 'nonagésimo'
}

_ORDINAL_HUNDREDS_MASC: Dict[int, str] = {
    100: 'centésimo', 200: 'ducentésimo', 300: 'trecentésimo',
    400: 'quadrigentésimo', 500: 'quingentésimo', 600: 'sexcentésimo',
    700: 'septingentésimo', 800: 'octingentésimo', 900: 'noningentésimo'
}

_ORDINAL_SCALES_MASC: Dict[Scale, Dict[PortugueseVariant, List[Tuple[int, str]]]] = {
    Scale.SHORT: {
        PortugueseVariant.BR: [
            (10 ** 21, "sextilionésimo"),
            (10 ** 18, "quintilionésimo"),
            (10 ** 15, "quadrilionésimo"),
            (10 ** 12, "trilionésimo"),
            (10 ** 9, "bilionésimo"),
            (10 ** 6, "milionésimo"),
            (10 ** 3, "milésimo")
        ],
        PortugueseVariant.PT: [
            (10 ** 21, "sextilionésimo"),
            (10 ** 18, "quintilionésimo"),
            (10 ** 15, "quatrilionésimo"),
            (10 ** 12, "trilionésimo"),
            (10 ** 9, "bilionésimo"),
            (10 ** 6, "milionésimo"),
            (10 ** 3, "milésimo")
        ]
    },
    Scale.LONG: {
        PortugueseVariant.BR: [
            (10 ** 36, "sextilionésimo"),
            (10 ** 30, "quintilionésimo"),
            (10 ** 24, "quatrilionésimo"),
            (10 ** 18, "trilionésimo"),
            (10 ** 12, "bilionésimo"),
            (10 ** 6, "milionésimo"),
            (10 ** 3, "milésimo")
        ],
        PortugueseVariant.PT: [
            (10 ** 36, "sextilionésimo"),
            (10 ** 30, "quintilionésimo"),
            (10 ** 24, "quatrilionésimo"),
            (10 ** 18, "trilionésimo"),
            (10 ** 12, "bilionésimo"),
            (10 ** 6, "milionésimo"),
            (10 ** 3, "milésimo")
        ]
    }
}


# --- Scale Definitions ---
# Structure: (value, singular_name, plural_name)
# Ordered from largest to smallest.
_SCALES: Dict[Scale, Dict[PortugueseVariant, List[Tuple[int, str, str]]]] = {
    Scale.SHORT: {
        PortugueseVariant.BR: [
            (10 ** 21, "sextilhão", "sextilhões"),
            (10 ** 18, "quintilhão", "quintilhões"),
            (10 ** 15, "quadrilhão", "quadrilhões"),
            (10 ** 12, "trilhão", "trilhões"),
            (10 ** 9, "bilhão", "bilhões"),
            (10 ** 6, "milhão", "milhões"),
            (10 ** 3, "mil", "mil")
        ],
        PortugueseVariant.PT: [
            (10 ** 21, "sextilião", "sextiliões"),
            (10 ** 18, "quintilião", "quintiliões"),
            (10 ** 15, "quatrilião", "quatriliões"),
            (10 ** 12, "trilião", "triliões"),
            (10 ** 9, "bilião", "biliões"),
            (10 ** 6, "milhão", "milhões"),
            (10 ** 3, "mil", "mil")
        ]
    },
    Scale.LONG: {
        PortugueseVariant.BR: [
            (10 ** 36, "sextilhão", "sextilhões"),
            (10 ** 30, "quintilhão", "quintilhões"),
            (10 ** 24, "quatrilhão", "quatrilhões"),
            (10 ** 18, "trilhão", "trilhões"),
            (10 ** 12, "bilhão", "bilhões"),
            (10 ** 6, "milhão", "milhões"),
            (10 ** 3, "mil", "mil")
        ],
        PortugueseVariant.PT: [
            (10 ** 36, "sextilião", "sextiliões"),
            (10 ** 30, "quintilião", "quintiliões"),
            (10 ** 24, "quatrilião", "quatriliões"),
            (10 ** 18, "trilião", "triliões"),
            (10 ** 12, "bilião", "biliões"),
            (10 ** 6, "milhão", "milhões"),
            (10 ** 3, "mil", "mil")
        ]
    }
}

# Mapping of number words to their integer values. This is dynamically built
# from the base dictionaries to ensure consistency and variant support.
_NUMBERS_BASE = {
    **_FEMALE_NUMS,
    **{v: k for k, v in _UNITS.items()},
    **{v: k for k, v in _TENS_PT.items()},
    **{v: k for k, v in _TENS_BR.items()},
    **{v: k for k, v in _HUNDREDS.items()},
    "cento": 100
}

def get_number_map(scale: Scale = Scale.LONG,
                   variant: PortugueseVariant = PortugueseVariant.PT):
    """
    Return a dictionary mapping Portuguese number words, including scale names, to their integer values for the specified scale and language variant.

    Parameters:
        scale (Scale): The numerical scale to use (short or long).
        variant (PortugueseVariant): The Portuguese language variant (Brazilian or European).

    Returns:
        dict: Mapping of Portuguese number words (units, tens, hundreds, scale names) to their corresponding integer values.
    """
    return {
        **_NUMBERS_BASE,
        **{s_name: val for val, s_name, _ in _SCALES[scale][variant]},
        **{p_name: val for val, _, p_name in _SCALES[scale][variant]}
    }

_NUMBERS_BR = get_number_map(Scale.SHORT, PortugueseVariant.BR)
_NUMBERS_PT = get_number_map(Scale.LONG, PortugueseVariant.PT)

_ORDINAL_WORDS_MASC = {
    **{v: k for k, v in _ORDINAL_UNITS_MASC.items()},
    **{v: k for k, v in _ORDINAL_TENS_MASC.items()},
    **{v: k for k, v in _ORDINAL_HUNDREDS_MASC.items()},
    **{s_name: val for val, s_name in _ORDINAL_SCALES_MASC[Scale.SHORT][PortugueseVariant.BR]},
}


def _swap_gender(word: str, gender: GrammaticalGender) -> str:
    """
    Convert a Portuguese word between masculine and feminine grammatical gender by adjusting its ending.

    Parameters:
        word (str): The word to convert.
        gender (GrammaticalGender): The target grammatical gender.

    Returns:
        str: The word with its ending swapped to match the specified gender, if applicable; otherwise, the original word.
    """
    if gender == GrammaticalGender.FEMININE and word.endswith('o'):
        return word[:-1] + 'a'
    elif gender == GrammaticalGender.MASCULINE and word.endswith('a'):
        return word[:-1] + 'o'
    elif gender == GrammaticalGender.FEMININE and word.endswith('os'):
        return word[:-2] + 'as'
    elif gender == GrammaticalGender.MASCULINE and word.endswith('as'):
        return word[:-2] + 'os'
    return word


def _pronounce_up_to_999(
        n: int,
        variant: PortugueseVariant = PortugueseVariant.BR
) -> str:
    """
    Returns the Portuguese cardinal pronunciation of an integer from 0 to 999, using the specified language variant.

    Parameters:
        n (int): Integer to pronounce (must be between 0 and 999).
        variant (PortugueseVariant, optional): Portuguese variant (Brazilian or European). Defaults to Brazilian.

    Returns:
        str: The number pronounced in Portuguese words.

    Raises:
        ValueError: If n is not in the range 0 to 999.
    """
    if not 0 <= n <= 999:
        raise ValueError("Number must be between 0 and 999.")
    if n == 0:
        return "zero"
    if n == 100:
        return "cem"

    parts = []
    tens_map = _TENS_BR if variant == PortugueseVariant.BR else _TENS_PT

    # Hundreds
    if n >= 100:
        hundred = n // 100 * 100
        parts.append("cento" if hundred == 100 else _HUNDREDS[hundred])
        n %= 100
        if n > 0:
            parts.append("e")

    # Tens and Units
    if n > 0:
        if n < 20:
            parts.append(tens_map.get(n) or _UNITS.get(n, ""))
        else:
            ten = n // 10 * 10
            unit = n % 10
            parts.append(tens_map[ten])
            if unit > 0:
                parts.append("e")
                parts.append(_UNITS[unit])

    return " ".join(parts)


def _pronounce_ordinal_up_to_999(
        n: int,
        gender: GrammaticalGender = GrammaticalGender.MASCULINE,
        variant: PortugueseVariant = PortugueseVariant.PT
) -> str:
    """
    Returns the Portuguese ordinal word for an integer between 0 and 999, adjusting for grammatical gender and language variant.

    Parameters:
        n (int): The integer to convert (must be between 0 and 999).

    Returns:
        str: The ordinal representation of the number in Portuguese.

    Raises:
        ValueError: If n is not between 0 and 999.
    """
    if not 0 <= n <= 999:
        raise ValueError("Number must be between 0 and 999.")
    if n == 0:
        return "zero"

    parts = []

    # Handle hundreds
    if n >= 100:
        hundred_val = n // 100 * 100
        hundred_word_masc = _ORDINAL_HUNDREDS_MASC.get(hundred_val)
        if hundred_word_masc:
            parts.append(_swap_gender(hundred_word_masc, gender))
        n %= 100

    # Handle tens and units
    if n > 0:
        # Ordinal numbers don't use 'e' as a separator
        if n % 10 == 0 and n > 10:
            tens_word_masc = _ORDINAL_TENS_MASC[n]
            parts.append(_swap_gender(tens_word_masc, gender))
        elif n < 10:
            units_word_masc = _ORDINAL_UNITS_MASC[n]
            parts.append(_swap_gender(units_word_masc, gender))
        elif n < 20:
            tens_word_masc = _ORDINAL_TENS_MASC[10]
            units_word_masc = _ORDINAL_UNITS_MASC[n - 10]
            parts.append(f"{_swap_gender(tens_word_masc, gender)} {_swap_gender(units_word_masc, gender)}")
        else:
            tens_word_masc = _ORDINAL_TENS_MASC[n // 10 * 10]
            units_word_masc = _ORDINAL_UNITS_MASC[n % 10]
            parts.append(f"{_swap_gender(tens_word_masc, gender)} {_swap_gender(units_word_masc, gender)}")

    return " ".join(parts)


def pronounce_ordinal_pt(
        number: Union[int, float],
        gender: GrammaticalGender = GrammaticalGender.MASCULINE,
        scale: Scale = Scale.LONG,
        variant: PortugueseVariant = PortugueseVariant.PT
) -> str:
    """
    Return the ordinal pronunciation of a number in Portuguese, supporting grammatical gender, scale (short or long), and language variant (Brazilian or European Portuguese).

    Parameters:
        number (int or float): The number to pronounce as an ordinal.
        gender (GrammaticalGender, optional): The grammatical gender for the ordinal form (masculine or feminine).
        scale (Scale, optional): The numerical scale to use (short or long).
        variant (PortugueseVariant, optional): The Portuguese variant (Brazilian or European).

    Returns:
        str: The ordinal pronunciation of the number in Portuguese.

    Raises:
        TypeError: If `number` is not an int or float.
    """
    if not isinstance(number, (int, float)):
        raise TypeError("Number must be an int or float.")
    if number == 0:
        return "zero"

    if number < 0:
        return f"menos {pronounce_ordinal_pt(abs(number), gender, scale, variant)}"

    n = int(number)
    if n < 1000:
        return _pronounce_ordinal_up_to_999(n, gender, variant)

    ordinal_scale_defs = _ORDINAL_SCALES_MASC[scale][variant]

    # Find the largest scale that fits the number
    for scale_val, s_name in ordinal_scale_defs:
        if n >= scale_val:
            break

    count = n // scale_val
    remainder = n % scale_val

    # Special case for "milésimo" and other large scales where 'um' is not needed
    if count == 1 and scale_val >= 1000:
        count_str = _swap_gender(s_name, gender)
    else:
        # Pronounce the 'count' part of the number and the scale word
        count_pronunciation = pronounce_number_pt(count, scale=scale, variant=variant)
        scale_word_masc = s_name
        scale_word = _swap_gender(scale_word_masc, gender)
        count_str = f"{count_pronunciation} {scale_word}"

    # If there's no remainder, we're done
    if remainder == 0:
        return count_str

    # Pronounce the remainder and join
    remainder_str = pronounce_ordinal_pt(remainder, gender, scale, variant)

    return f"{count_str} {remainder_str}"


def is_fractional_pt(
        input_str: str
) -> Union[float, bool]:
    """
    Checks if the input string corresponds to a recognized Portuguese fractional word.

    Returns:
        The fractional value as a float if recognized (e.g., 0.5 for "meio" or "meia"); otherwise, False.
    """
    input_str = input_str.lower().strip()
    fraction_map = _FRACTION_STRING_PT

    # Handle plural forms
    if input_str.endswith('s') and input_str not in fraction_map.values():
        input_str = input_str[:-1]

    # Handle "meio" vs "meia"
    if input_str == "meia":
        input_str = "meio"

    # Use a dynamic lookup instead of a hardcoded list
    for den, word in fraction_map.items():
        # Handle cases like "onze avos", so we check for the whole word
        if input_str == word:
            return 1.0 / den

    # Special case for "meia" as a female form of "meio" (1/2)
    if input_str in ["meia", "meio"]:
        return 0.5

    return False


def is_ordinal_pt(input_str: str) -> bool:
    """
    Determine if a string is a Portuguese ordinal word.

    Returns:
        bool: True if the input string is recognized as a Portuguese ordinal, otherwise False.
    """
    input_str = _swap_gender(input_str, GrammaticalGender.MASCULINE)
    return input_str in _ORDINAL_WORDS_MASC


def extract_number_pt(
        text: str,
        ordinals: bool = False,
        scale: Scale = Scale.LONG,
        variant: PortugueseVariant = PortugueseVariant.PT
) -> Union[int, float, bool]:
    """
    Extracts a numeric value from a Portuguese text phrase, supporting cardinals, ordinals, fractions, and large scales.

    Parameters:
        text (str): The input phrase potentially containing a number.
        ordinals (bool): If True, recognizes ordinal words as numbers.
        scale (Scale): Specifies whether to use the short or long numerical scale.
        variant (PortugueseVariant): Specifies the Portuguese language variant (BR or PT).

    Returns:
        int or float: The extracted number if found; otherwise, False.
    """
    numbers_map = get_number_map(scale, variant)
    scales_map = _SCALES[scale][variant]

    clean_text = text.lower().replace('-', ' ')
    tokens = [t for t in clean_text.split() if t != "e"]

    result = 0
    current_number = 0
    number_consumed = False

    for i, token in enumerate(tokens):
        if token is None:
            continue # consumed in previous idx
        next_token = tokens[i+1] if i < len(tokens) - 1 else None
        next_digit = numbers_map.get(next_token) if next_token else None
        val = numbers_map.get(token)
        if val is not None:
            if next_digit and  next_digit > val:
                tokens[i+1] = None
                current_number += val * next_digit
            else:
                current_number += val
        elif ordinals and is_ordinal_pt(token):
            token = _swap_gender(token, GrammaticalGender.MASCULINE)
            current_number += _ORDINAL_WORDS_MASC[token]
        elif is_fractional_pt(token):
            fraction = is_fractional_pt(token)
            result += current_number + fraction
            current_number = 0
            number_consumed = True
        else:
            # Handle large scales like milhão, bilhão
            found_scale = False
            for scale_val, singular, plural in scales_map:
                if token == singular or token == plural:
                    if current_number == 0:
                        current_number = 1
                    result += current_number * scale_val
                    current_number = 0
                    found_scale = True
                    number_consumed = True
                    break
            if not found_scale:
                if token in ["ponto", "virgula", "vírgula", ".", ","]:
                    decimal_str = ''.join(
                        str(numbers_map.get(t, '')) for t in tokens[i+1:]
                        if t in numbers_map
                    )
                    if decimal_str:
                        result += current_number + float(f"0.{decimal_str}")
                        number_consumed = True
                    current_number = 0
                    break

    if not number_consumed:
        result += current_number

    return result if result > 0 else False



def pronounce_number_pt(
        number: Union[int, float],
        places: int = 5,
        scale: Scale = Scale.LONG,
        variant: PortugueseVariant = PortugueseVariant.PT,
        ordinals: bool = False,
        digits: DigitPronunciation = DigitPronunciation.FULL_NUMBER,
        gender: GrammaticalGender = GrammaticalGender.MASCULINE
) -> str:
    """
    Return the full Portuguese pronunciation of a number, supporting cardinal and ordinal forms, decimals, grammatical gender, and both Brazilian and European Portuguese variants.
    
    Parameters:
        number (int or float): The number to pronounce.
        places (int): Number of decimal places to include for floats.
        scale (Scale): Numerical scale to use (short or long).
        variant (PortugueseVariant): Portuguese language variant for pronunciation.
        ordinals (bool): If True, pronounce as an ordinal number.
        digits (DigitPronunciation): Determines whether decimal parts are pronounced as a whole number or digit by digit.
        gender (GrammaticalGender): Grammatical gender for ordinal numbers.
    
    Returns:
        str: The number expressed as a Portuguese phrase.
    """
    if not isinstance(number, (int, float)):
        raise TypeError("Number must be an int or float.")

    if ordinals:
        return pronounce_ordinal_pt(number, gender, scale, variant)

    if number == 0:
        return "zero"

    if number < 0:
        return f"menos {pronounce_number_pt(abs(number), places, scale, variant)}"

    # Handle decimals
    if isinstance(number, float):
        integer_part = int(number)
        decimal_part_str = f"{number:.{places}f}".split('.')[1].rstrip("0")

        # Handle cases where the decimal part rounds to zero
        if decimal_part_str and int(decimal_part_str) == 0:
            return pronounce_number_pt(integer_part, places, scale, variant)

        int_pronunciation = pronounce_number_pt(integer_part, places, scale, variant)

        decimal_pronunciation_parts = []
        #  pronounce decimals either as a whole number or digit by digit
        if decimal_part_str:
            if digits == DigitPronunciation.FULL_NUMBER:
                decimal_pronunciation_parts.append(_pronounce_up_to_999(int(decimal_part_str[:3]), variant))
            else:
                for digit in decimal_part_str:
                    decimal_pronunciation_parts.append(_pronounce_up_to_999(int(digit), variant))

        decimal_pronunciation = " ".join(decimal_pronunciation_parts) or "zero"
        decimal_word = "vírgula"
        return f"{int_pronunciation} {decimal_word} {decimal_pronunciation}"

    # --- Integer Pronunciation Logic ---
    n = int(number)

    # Base case for recursion: numbers less than 1000
    if n < 1000:
        return _pronounce_up_to_999(n, variant)

    scale_definitions = _SCALES[scale][variant]

    # Find the largest scale that fits the number
    for scale_val, s_name, p_name in scale_definitions:
        if n >= scale_val:
            break

    count = n // scale_val
    remainder = n % scale_val

    # Pronounce the 'count' part of the number
    scale_word = s_name if count == 1 else p_name
    if count == 1 and scale_word == "mil":
        count_str = scale_word
    else:
        count_pronunciation = pronounce_number_pt(count, places, scale, variant)
        count_str = f"{count_pronunciation} {scale_word}"

    # If there's no remainder, we're done
    if remainder == 0:
        return count_str

    # Pronounce the remainder and join with the correct conjunction
    remainder_str = pronounce_number_pt(remainder, places, scale, variant)

    # Conjunction logic: add "e" if the remainder is the last group and is
    # less than 100 or a multiple of 100.
    if remainder < 100 or (remainder < 1000 and remainder % 100 == 0):
        return f"{count_str} e {remainder_str}"
    else:
        return f"{count_str} {remainder_str}"


def numbers_to_digits_pt(
        utterance: str,
        scale: Scale = Scale.LONG,
        variant: PortugueseVariant = PortugueseVariant.PT
) -> str:
    """
    Converts written Portuguese numbers in a text string to their digit equivalents, preserving all other text.

    Identifies spans of number words (including the joiner "e"), extracts their numeric values, and replaces them with digit strings. Non-number words and context are left unchanged.

    Parameters:
        utterance (str): Input text possibly containing written Portuguese numbers.
        scale (Scale, optional): Numerical scale (short or long) to interpret large numbers. Defaults to Scale.LONG.
        variant (PortugueseVariant, optional): Portuguese language variant (BR or PT). Defaults to PortugueseVariant.PT.

    Returns:
        str: The input text with written numbers replaced by their digit representations.
    """
    words = tokenize(utterance)
    output = []
    i = 0
    NUMBERS = get_number_map(scale, variant)
    while i < len(words):
        # Look for the start of a number span
        if words[i] in NUMBERS:
            # Start a new span
            number_span_words = []
            j = i
            # Continue the span as long as we find number words or the joiner 'e'
            while j < len(words) and (words[j] in NUMBERS or words[j] == "e"):
                number_span_words.append(words[j])
                j += 1

            # Form the phrase from the span and extract the number value
            phrase = " ".join(number_span_words)
            number_val = extract_number_pt(phrase, variant=variant)

            if number_val is not False:
                # If a valid number is found, add its digit representation to the output
                output.append(str(number_val))
                # Advance the main index 'i' past the entire span
                i = j
            else:
                # If the span doesn't form a valid number, treat the first word as non-numeric
                # and move to the next word. This handles cases like "e" at the beginning of a sentence.
                output.append(words[i])
                i += 1
        else:
            # If the current word is not a number word, add it to the output
            # and move to the next word
            output.append(words[i])
            i += 1

    return " ".join(output)


def tokenize(utterance: str) -> List[str]:
    """
    Splits a Portuguese text string into a list of tokens, separating words and punctuation.

    Returns:
        A list of tokens, where each token is a word or punctuation mark from the input string.
    """
    # Split things like 12%
    utterance = re.sub(r"([0-9]+)([\%])", r"\1 \2", utterance)
    # Split things like #1
    utterance = re.sub(r"(\#)([0-9]+\b)", r"\1 \2", utterance)
    # Split things like amo-te, but preserve numbers like 1-2
    utterance = re.sub(r"([a-zA-Z]+)(-)([a-zA-Z]+\b)", r"\1 \2 \3",
                       utterance)

    tokens = utterance.split()
    # Remove a trailing hyphen if it's the last token
    if tokens and tokens[-1] == '-':
        tokens = tokens[:-1]

    return tokens


def pronounce_fraction_pt(word: str,
                          scale: Scale = Scale.LONG,
                          variant: PortugueseVariant = PortugueseVariant.PT) -> str:
    """
    Return the Portuguese pronunciation of a fraction given as a string (e.g., "1/2").

    The numerator is pronounced as a cardinal number, and the denominator as an ordinal or fraction name, pluralized if appropriate. For denominators not in the known fraction list, the denominator is pronounced as a cardinal number followed by "avos" if plural.

    Parameters:
        word (str): Fraction in the form "numerator/denominator" (e.g., "3/4").

    Returns:
        str: The Portuguese pronunciation of the fraction.
    """
    n1, n2 = word.split("/")
    n1_int, n2_int = int(n1), int(n2)

    # Pronounce the denominator (second number) as an ordinal, and pluralize it if needed.
    if n2_int in _FRACTION_STRING_PT:
        denom = _FRACTION_STRING_PT[n2_int]
        if n1_int != 1:
            denom += "s" # plural
    else:
        # For other numbers
        denom = pronounce_number_pt(n2_int, scale=scale, variant=variant)
        if n1_int > 1:  # plural
            denom += " avos"

    # Pronounce the numerator (first number) as a cardinal.
    num = pronounce_number_pt(n1_int, scale=scale, variant=variant)
    return f"{num} {denom}"


if __name__ == "__main__":
    print("--- Testing Pronunciation (Short Scale, BR Variant) ---")
    print(f"1,234,567: {pronounce_number_pt(1_234_567, scale=Scale.SHORT, variant=PortugueseVariant.BR)}")
    print(f"1,000,000,000: {pronounce_number_pt(1_000_000_000, scale=Scale.SHORT, variant=PortugueseVariant.BR)}")
    print(f"16: {pronounce_number_pt(16, variant=PortugueseVariant.BR)}")

    print("\n--- Testing Pronunciation (Long Scale, PT Variant) ---")
    print(f"1,000,000: {pronounce_number_pt(1_000_000, scale=Scale.LONG, variant=PortugueseVariant.PT)}")
    print(f"1,000,100: {pronounce_number_pt(1_000_100, scale=Scale.LONG, variant=PortugueseVariant.PT)}")
    print(f"1,000,000,000: {pronounce_number_pt(1_000_000_000, scale=Scale.LONG, variant=PortugueseVariant.PT)}")
    print(f"1,000,000,000,000: {pronounce_number_pt(1_000_000_000_000, scale=Scale.LONG, variant=PortugueseVariant.PT)}")
    print(f"2,500,000,000: {pronounce_number_pt(2_500_000_000, scale=Scale.LONG, variant=PortugueseVariant.PT)}")
    print(f"2,500,123,456: {pronounce_number_pt(2_500_123_456, scale=Scale.LONG, variant=PortugueseVariant.PT)}")
    print(f"16: {pronounce_number_pt(16, variant=PortugueseVariant.PT)}")

    print("\n--- Testing Edge Cases ---")
    print(f"-123.45: {pronounce_number_pt(-123.45)}")
    print(f"10.05: {pronounce_number_pt(10.05)}")
    print(f"2000: {pronounce_number_pt(2000)}")
    print(f"2001: {pronounce_number_pt(2001)}")
    print(f"123.456789: {pronounce_number_pt(123.456789)}")

    print("\n--- Testing Ordinal Pronunciation ---")
    print(f"1st (masculine): {pronounce_number_pt(1, ordinals=True, gender=GrammaticalGender.MASCULINE)}")
    print(f"1st (feminine): {pronounce_number_pt(1, ordinals=True, gender=GrammaticalGender.FEMININE)}")
    print(f"23rd (masculine): {pronounce_number_pt(23, ordinals=True)}")
    print(f"23rd (feminine): {pronounce_number_pt(23, ordinals=True, gender=GrammaticalGender.FEMININE)}")
    print(f"100th: {pronounce_number_pt(100, ordinals=True)}")
    print(f"101st: {pronounce_number_pt(101, ordinals=True)}")
    print(f"1000th: {pronounce_number_pt(1000, ordinals=True)}")
    print(f"1,000,000th: {pronounce_number_pt(1_000_000, ordinals=True)}")
    print(f"1,000,000,000,000th (PT, long): {pronounce_number_pt(1_000_000_000_000, ordinals=True, variant=PortugueseVariant.PT, scale=Scale.LONG)}")

    print("\n--- Testing numbers_to_digits_pt (BR) ---")
    print(f"'quinhentos e cinquenta' -> '{numbers_to_digits_pt('quinhentos e cinquenta')}'")
    print(f"'um milhão' -> '{numbers_to_digits_pt('um milhão')}'")
    print(f"'dezesseis' -> '{numbers_to_digits_pt('dezesseis')}'")
    print(f"'há duzentos e cinquenta carros' -> '{numbers_to_digits_pt('há duzentos e cinquenta carros')}'")

    print("\n--- Testing numbers_to_digits_pt (PT) ---")
    print(f"'quinhentos e cinquenta' -> '{numbers_to_digits_pt('quinhentos e cinquenta', variant=PortugueseVariant.PT)}'")
    print(f"'um milhão' -> '{numbers_to_digits_pt('um milhão', variant=PortugueseVariant.PT)}'")
    print(f"'dezasseis' -> '{numbers_to_digits_pt('dezasseis', variant=PortugueseVariant.PT)}'")

    print("\n--- Testing Ordinal Extraction ---")
    print(f"'o segundo carro' -> {extract_number_pt('o segundo carro', ordinals=True)}")
    print(f"'primeiro lugar' -> {extract_number_pt('primeiro lugar', ordinals=True)}")
    print(f"'o milésimo dia' -> {extract_number_pt('o milésimo dia', ordinals=True)}")
    print(f"'a milésima vez' -> {extract_number_pt('a milésima vez', ordinals=True)}")
    print(f"'a primeira vez' -> {extract_number_pt('a primeira vez', ordinals=True)}")
    print(f"'a sexagésima quarta vez' -> {extract_number_pt('a sexagésima quarta vez', ordinals=True)}")

    print("\n--- Testing Cardinal Extraction ---")
    print(f"'um' -> {extract_number_pt('um')}")
    print(f"'uma' -> {extract_number_pt('uma')}")
    print(f"'vinte e um' -> {extract_number_pt('vinte e um')}")
    print(f"'um milhão' -> {extract_number_pt('um milhão')}")
    print(f"'dois milhões e quinhentos' -> {extract_number_pt('dois milhões e quinhentos')}")
    print(f"'mil e vinte e três' -> {extract_number_pt('mil e vinte e três')}")
    print(f"'trinta e cinco vírgula quatro' -> {extract_number_pt('trinta e cinco vírgula quatro')}")


    print("\n--- Testing Fractions ---")
    print(f"1/2: {pronounce_fraction_pt('1/2')}")
    print(f"2/2: {pronounce_fraction_pt('2/2')}")
    print(f"5/2: {pronounce_fraction_pt('5/2')}")
    print(f"5/3: {pronounce_fraction_pt('5/3')}")
    print(f"5/4: {pronounce_fraction_pt('5/4')}")
    print(f"7/5: {pronounce_fraction_pt('7/5')}")
    print(f"0/20: {pronounce_fraction_pt('0/20')}")




#
# Copyright 2019 Mycroft AI Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#

from collections import OrderedDict
from math import floor

from ovos_number_parser.util import convert_to_mixed_fraction, is_numeric, look_for_fractions, Token, \
    ReplaceableNumber, tokenize, partition_list, invert_dict

_ARTICLES_NL = {'de', 'het'}

_NUM_STRING_NL = {
    0: 'nul',
    1: 'een',
    2: 'twee',
    3: 'drie',
    4: 'vier',
    5: 'vijf',
    6: 'zes',
    7: 'zeven',
    8: 'acht',
    9: 'negen',
    10: 'tien',
    11: 'elf',
    12: 'twaalf',
    13: 'dertien',
    14: 'veertien',
    15: 'vijftien',
    16: 'zestien',
    17: 'zeventien',
    18: 'achttien',
    19: 'negentien',
    20: 'twintig',
    30: 'dertig',
    40: 'veertig',
    50: 'vijftig',
    60: 'zestig',
    70: 'zeventig',
    80: 'tachtig',
    90: 'negentig'
}

_FRACTION_STRING_NL = {
    2: 'half',
    3: 'derde',
    4: 'vierde',
    5: 'vijfde',
    6: 'zesde',
    7: 'zevende',
    8: 'achtste',
    9: 'negende',
    10: 'tiende',
    11: 'elfde',
    12: 'twaalfde',
    13: 'dertiende',
    14: 'veertiende',
    15: 'vijftiende',
    16: 'zestiende',
    17: 'zeventiende',
    18: 'achttiende',
    19: 'negentiende',
    20: 'twintigste'
}

_LONG_SCALE_NL = OrderedDict([
    (100, 'honderd'),
    (1000, 'duizend'),
    (1000000, 'miljoen'),
    (1e12, "biljoen"),
    (1e18, 'triljoen'),
    (1e24, "quadriljoen"),
    (1e30, "quintillion"),
    (1e36, "sextillion"),
    (1e42, "septillion"),
    (1e48, "octillion"),
    (1e54, "nonillion"),
    (1e60, "decillion"),
    (1e66, "undecillion"),
    (1e72, "duodecillion"),
    (1e78, "tredecillion"),
    (1e84, "quattuordecillion"),
    (1e90, "quinquadecillion"),
    (1e96, "sedecillion"),
    (1e102, "septendecillion"),
    (1e108, "octodecillion"),
    (1e114, "novendecillion"),
    (1e120, "vigintillion"),
    (1e306, "unquinquagintillion"),
    (1e312, "duoquinquagintillion"),
    (1e336, "sesquinquagintillion"),
    (1e366, "unsexagintillion")
])

_SHORT_SCALE_NL = OrderedDict([
    (100, 'honderd'),
    (1000, 'duizend'),
    (1000000, 'miljoen'),
    (1e9, "miljard"),
    (1e12, 'biljoen'),
    (1e15, "quadrillion"),
    (1e18, "quintiljoen"),
    (1e21, "sextiljoen"),
    (1e24, "septiljoen"),
    (1e27, "octiljoen"),
    (1e30, "noniljoen"),
    (1e33, "deciljoen"),
    (1e36, "undeciljoen"),
    (1e39, "duodeciljoen"),
    (1e42, "tredeciljoen"),
    (1e45, "quattuordeciljoen"),
    (1e48, "quinquadeciljoen"),
    (1e51, "sedeciljoen"),
    (1e54, "septendeciljoen"),
    (1e57, "octodeciljoen"),
    (1e60, "novendeciljoen"),
    (1e63, "vigintiljoen"),
    (1e66, "unvigintiljoen"),
    (1e69, "uuovigintiljoen"),
    (1e72, "tresvigintiljoen"),
    (1e75, "quattuorvigintiljoen"),
    (1e78, "quinquavigintiljoen"),
    (1e81, "qesvigintiljoen"),
    (1e84, "septemvigintiljoen"),
    (1e87, "octovigintiljoen"),
    (1e90, "novemvigintiljoen"),
    (1e93, "trigintiljoen"),
    (1e96, "untrigintiljoen"),
    (1e99, "duotrigintiljoen"),
    (1e102, "trestrigintiljoen"),
    (1e105, "quattuortrigintiljoen"),
    (1e108, "quinquatrigintiljoen"),
    (1e111, "sestrigintiljoen"),
    (1e114, "septentrigintiljoen"),
    (1e117, "octotrigintiljoen"),
    (1e120, "noventrigintiljoen"),
    (1e123, "quadragintiljoen"),
    (1e153, "quinquagintiljoen"),
    (1e183, "sexagintiljoen"),
    (1e213, "septuagintiljoen"),
    (1e243, "octogintiljoen"),
    (1e273, "nonagintiljoen"),
    (1e303, "centiljoen"),
    (1e306, "uncentiljoen"),
    (1e309, "duocentiljoen"),
    (1e312, "trescentiljoen"),
    (1e333, "decicentiljoen"),
    (1e336, "undecicentiljoen"),
    (1e363, "viginticentiljoen"),
    (1e366, "unviginticentiljoen"),
    (1e393, "trigintacentiljoen"),
    (1e423, "quadragintacentiljoen"),
    (1e453, "quinquagintacentiljoen"),
    (1e483, "sexagintacentiljoen"),
    (1e513, "septuagintacentiljoen"),
    (1e543, "ctogintacentiljoen"),
    (1e573, "nonagintacentiljoen"),
    (1e603, "ducentiljoen"),
    (1e903, "trecentiljoen"),
    (1e1203, "quadringentiljoen"),
    (1e1503, "quingentiljoen"),
    (1e1803, "sescentiljoen"),
    (1e2103, "septingentiljoen"),
    (1e2403, "octingentiljoen"),
    (1e2703, "nongentiljoen"),
    (1e3003, "milliniljoen")
])

_ORDINAL_STRING_BASE_NL = {
    1: 'eerste',
    2: 'tweede',
    3: 'derde',
    4: 'vierde',
    5: 'vijfde',
    6: 'zesde',
    7: 'zevende',
    8: 'achtste',
    9: 'negende',
    10: 'tiende',
    11: 'elfde',
    12: 'twaalfde',
    13: 'dertiende',
    14: 'veertiende',
    15: 'vijftiende',
    16: 'zestiende',
    17: 'zeventiende',
    18: 'achttiende',
    19: 'negentiende',
    20: 'twintigste',
    30: 'dertigste',
    40: "veertigste",
    50: "vijftigste",
    60: "zestigste",
    70: "zeventigste",
    80: "tachtigste",
    90: "negentigste",
    10e3: "honderdste",
    1e3: "duizendste"
}

_SHORT_ORDINAL_STRING_NL = {
    1e6: "miloenste",
    1e9: "miljardste",
    1e12: "biljoenste",
    1e15: "biljardste",
    1e18: "triljoenste",
    1e21: "trijardste",
    1e24: "quadriljoenste",
    1e27: "quadriljardste",
    1e30: "quintiljoenste",
    1e33: "quintiljardste"
    # TODO > 1e-33
}
_SHORT_ORDINAL_STRING_NL.update(_ORDINAL_STRING_BASE_NL)

_LONG_ORDINAL_STRING_NL = {
    1e6: "miloenste",
    1e9: "miljardste",
    1e12: "biljoenste",
    1e15: "biljardste",
    1e18: "triljoenste",
    1e21: "trijardste",
    1e24: "quadriljoenste",
    1e27: "quadriljardste",
    1e30: "quintiljoenste",
    1e33: "quintiljardste"
    # TODO > 1e60
}
_LONG_ORDINAL_STRING_NL.update(_ORDINAL_STRING_BASE_NL)

# negate next number (-2 = 0 - 2)
_NEGATIVES_NL = {"min", "minus"}

# sum the next number (twenty two = 20 + 2)
_SUMS_NL = {'twintig', '20', 'dertig', '30', 'veertig', '40', 'vijftig', '50',
            'zestig', '60', 'zeventig', '70', 'techtig', '80', 'negentig',
            '90'}

_MULTIPLIES_LONG_SCALE_NL = set(_LONG_SCALE_NL.values())

_MULTIPLIES_SHORT_SCALE_NL = set(_SHORT_SCALE_NL.values())

# split sentence parse separately and sum ( 2 and a half = 2 + 0.5 )
_FRACTION_MARKER_NL = {"en"}

# decimal marker ( 1 point 5 = 1 + 0.5)
_DECIMAL_MARKER_NL = {"komma", "punt"}

_STRING_NUM_NL = invert_dict(_NUM_STRING_NL)
_STRING_NUM_NL.update({
    "half": 0.5,
    "driekwart": 0.75,
    "anderhalf": 1.5,
    "paar": 2
})

_STRING_SHORT_ORDINAL_NL = invert_dict(_SHORT_ORDINAL_STRING_NL)
_STRING_LONG_ORDINAL_NL = invert_dict(_LONG_ORDINAL_STRING_NL)

_MONTHS_NL = ['januari', 'februari', 'maart', 'april', 'mei', 'juni',
              'juli', 'augustus', 'september', 'oktober', 'november',
              'december']

_NUM_STRING_NL = {
    0: 'nul',
    1: 'één',
    2: 'twee',
    3: 'drie',
    4: 'vier',
    5: 'vijf',
    6: 'zes',
    7: 'zeven',
    8: 'acht',
    9: 'negen',
    10: 'tien',
    11: 'elf',
    12: 'twaalf',
    13: 'dertien',
    14: 'veertien',
    15: 'vijftien',
    16: 'zestien',
    17: 'zeventien',
    18: 'achttien',
    19: 'negentien',
    20: 'twintig',
    30: 'dertig',
    40: 'veertig',
    50: 'vijftig',
    60: 'zestig',
    70: 'zeventig',
    80: 'tachtig',
    90: 'negentig',
    100: 'honderd'
}

# Dutch uses "long scale" https://en.wikipedia.org/wiki/Long_and_short_scales
# Currently, numbers are limited to 1000000000000000000000000,
# but _NUM_POWERS_OF_TEN can be extended to include additional number words


_NUM_POWERS_OF_TEN = [
    '', 'duizend', 'miljoen', 'miljard', 'biljoen', 'biljard', 'triljoen',
    'triljard'
]

# Numbers below 1 million are written in one word in dutch, yielding very
# long words
# In some circumstances it may better to seperate individual words
# Set _EXTRA_SPACE_NL=" " for separating numbers below 1 million (
# orthographically incorrect)
# Set _EXTRA_SPACE_NL="" for correct spelling, this is standard

# _EXTRA_SPACE_NL = " "
_EXTRA_SPACE_NL = ""


def nice_number_nl(number, speech=True, denominators=range(1, 21)):
    """ Dutch helper for nice_number
    This function formats a float to human understandable functions. Like
    4.5 becomes "4 einhalb" for speech and "4 1/2" for text
    Args:
        number (int or float): the float to format
        speech (bool): format for speech (True) or display (False)
        denominators (iter of ints): denominators to use, default [1 .. 20]
    Returns:
        (str): The formatted string.
    """
    result = convert_to_mixed_fraction(number, denominators)
    if not result:
        # Give up, just represent as a 3 decimal number
        return str(round(number, 3)).replace(".", ",")
    whole, num, den = result
    if not speech:
        if num == 0:
            # TODO: Number grouping?  E.g. "1,000,000"
            return str(whole)
        else:
            return '{} {}/{}'.format(whole, num, den)
    if num == 0:
        return str(whole)
    den_str = _FRACTION_STRING_NL[den]
    if whole == 0:
        if num == 1:
            return_string = 'één {}'.format(den_str)
        else:
            return_string = '{} {}'.format(num, den_str)
    elif num == 1:
        return_string = '{} en één {}'.format(whole, den_str)
    else:
        return_string = '{} en {} {}'.format(whole, num, den_str)

    return return_string


def pronounce_number_nl(number, places=2, short_scale=True, scientific=False,
                        ordinals=False):
    """
    Convert a number to it's spoken equivalent

    For example, '5.2' would return 'five point two'

    Args:
        number(float or int): the number to pronounce (under 100)
        places(int): maximum decimal places to speak
        short_scale (bool) : use short (True) or long scale (False)
            https://en.wikipedia.org/wiki/Names_of_large_numbers
        scientific (bool): pronounce in scientific notation
        ordinals (bool): pronounce in ordinal form "first" instead of "one"
    Returns:
        (str): The pronounced number
    """

    # TODO short_scale, scientific and ordinals
    # currently ignored

    def pronounce_triplet_nl(num):
        result = ""
        num = floor(num)
        if num > 99:
            hundreds = floor(num / 100)
            if hundreds > 0:
                result += _NUM_STRING_NL[
                              hundreds] + _EXTRA_SPACE_NL + 'honderd' + _EXTRA_SPACE_NL
                num -= hundreds * 100
        if num == 0:
            result += ''  # do nothing
        elif num <= 20:
            result += _NUM_STRING_NL[num]  # + _EXTRA_SPACE_DA
        elif num > 20:
            ones = num % 10
            tens = num - ones
            if ones > 0:
                result += _NUM_STRING_NL[ones] + _EXTRA_SPACE_NL
                if tens > 0:
                    result += 'en' + _EXTRA_SPACE_NL
            if tens > 0:
                result += _NUM_STRING_NL[tens] + _EXTRA_SPACE_NL
        return result

    def pronounce_fractional_nl(num,
                                places):  # fixed number of places even with
        # trailing zeros
        result = ""
        place = 10
        while places > 0:  # doesn't work with 1.0001 and places = 2: int(
            # number*place) % 10 > 0 and places > 0:
            result += " " + _NUM_STRING_NL[int(num * place) % 10]
            if int(num * place) % 10 == 1:
                result += ''  # "1" is pronounced "eins" after the decimal
                # point
            place *= 10
            places -= 1
        return result

    def pronounce_whole_number_nl(num, scale_level=0):
        if num == 0:
            return ''

        num = floor(num)
        result = ''
        last_triplet = num % 1000

        if last_triplet == 1:
            if scale_level == 0:
                if result != '':
                    result += '' + 'één'
                else:
                    result += "één"
            elif scale_level == 1:
                result += 'één' + _EXTRA_SPACE_NL + 'duizend' + _EXTRA_SPACE_NL
            else:
                result += "één " + _NUM_POWERS_OF_TEN[scale_level] + ' '
        elif last_triplet > 1:
            result += pronounce_triplet_nl(last_triplet)
            if scale_level == 1:
                # result += _EXTRA_SPACE_DA
                result += 'duizend' + _EXTRA_SPACE_NL
            if scale_level >= 2:
                # if _EXTRA_SPACE_DA == '':
                #    result += " "
                result += " " + _NUM_POWERS_OF_TEN[scale_level] + ' '
            if scale_level >= 2:
                if scale_level % 2 == 0:
                    result += ""  # Miljioen
                result += ""  # Miljard, Miljoen

        num = floor(num / 1000)
        scale_level += 1
        return pronounce_whole_number_nl(num,
                                         scale_level) + result + ''

    result = ""
    if abs(number) >= 1000000000000000000000000:  # cannot do more than this
        return str(number)
    elif number == 0:
        return str(_NUM_STRING_NL[0])
    elif number < 0:
        return "min " + pronounce_number_nl(abs(number), places)
    else:
        if number == int(number):
            return pronounce_whole_number_nl(number)
        else:
            whole_number_part = floor(number)
            fractional_part = number - whole_number_part
            result += pronounce_whole_number_nl(whole_number_part)
            if places > 0:
                result += " komma"
                result += pronounce_fractional_nl(fractional_part, places)
            return result


def pronounce_ordinal_nl(number):
    """
    This function pronounces a number as an ordinal

    1 -> first
    2 -> second

    Args:
        number (int): the number to format
    Returns:
        (str): The pronounced number string.
    """
    ordinals = ["nulste", "eerste", "tweede", "derde", "vierde", "vijfde",
                "zesde", "zevende", "achtste"]
    # only for whole positive numbers including zero
    if number < 0 or number != int(number):
        return number
    if number < 4:
        return ordinals[number]
    if number < 8:
        return pronounce_number_nl(number) + "de"
    if number < 9:
        return pronounce_number_nl(number) + "ste"
    if number < 20:
        return pronounce_number_nl(number) + "de"
    return pronounce_number_nl(number) + "ste"


def numbers_to_digits_nl(text, short_scale=True, ordinals=False):
    """Convert words in a string into their equivalent numbers.
    Args:
        text str:
        short_scale boolean: True if short scale numbers should be used.
        ordinals boolean: True if ordinals (e.g. first, second, third) should
                          be parsed to their number values (1, 2, 3...)

    Returns:
        str
        The original text, with numbers subbed in where appropriate.
    """
    text = text.lower()
    tokens = tokenize(text)
    numbers_to_replace = \
        _extract_numbers_with_text_nl(tokens, short_scale, ordinals)
    numbers_to_replace.sort(key=lambda number: number.start_index)

    results = []
    for token in tokens:
        if not numbers_to_replace or \
                token.index < numbers_to_replace[0].start_index:
            results.append(token.word)
        else:
            if numbers_to_replace and \
                    token.index == numbers_to_replace[0].start_index:
                results.append(str(numbers_to_replace[0].value))
            if numbers_to_replace and \
                    token.index == numbers_to_replace[0].end_index:
                numbers_to_replace.pop(0)

    return ' '.join(results)


def _extract_numbers_with_text_nl(tokens, short_scale=True,
                                  ordinals=False, fractional_numbers=True):
    """Extract all numbers from a list of _Tokens, with the representing words.

    Args:
        [Token]: The tokens to parse.
        short_scale bool: True if short scale numbers should be used, False for
                          long scale. True by default.
        ordinals bool: True if ordinal words (first, second, third, etc) should
                       be parsed.
        fractional_numbers bool: True if we should look for fractions and
                                 decimals.

    Returns:
        [_ReplaceableNumber]: A list of tuples, each containing a number and a
                         string.
    """
    placeholder = "<placeholder>"  # inserted to maintain correct indices
    results = []
    while True:
        to_replace = \
            _extract_number_with_text_nl(tokens, short_scale,
                                         ordinals, fractional_numbers)

        if not to_replace:
            break

        results.append(to_replace)

        tokens = [
            t if not
            to_replace.start_index <= t.index <= to_replace.end_index
            else
            Token(placeholder, t.index) for t in tokens
        ]
    results.sort(key=lambda n: n.start_index)
    return results


def _extract_number_with_text_nl(tokens, short_scale=True,
                                 ordinals=False, fractional_numbers=True):
    """This function extracts a number from a list of _Tokens.

    Args:
        tokens str: the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers, third=3 instead of 1/3
        fractional_numbers (bool): True if we should look for fractions and
                                   decimals.
    Returns:
        _ReplaceableNumber
    """
    number, tokens = \
        _extract_number_with_text_nl_helper(tokens, short_scale,
                                            ordinals, fractional_numbers)
    while tokens and tokens[0].word in _ARTICLES_NL:
        tokens.pop(0)
    return ReplaceableNumber(number, tokens)


def _extract_number_with_text_nl_helper(tokens,
                                        short_scale=True, ordinals=False,
                                        fractional_numbers=True):
    """Helper for _extract_number_with_text_nl.

    This contains the real logic for parsing, but produces
    a result that needs a little cleaning (specific, it may
    contain leading articles that can be trimmed off).

    Args:
        tokens [Token]:
        short_scale boolean:
        ordinals boolean:
        fractional_numbers boolean:

    Returns:
        int or float, [_Tokens]
    """
    if fractional_numbers:
        fraction, fraction_text = \
            _extract_fraction_with_text_nl(tokens, short_scale, ordinals)
        if fraction:
            return fraction, fraction_text

        decimal, decimal_text = \
            _extract_decimal_with_text_nl(tokens, short_scale, ordinals)
        if decimal:
            return decimal, decimal_text

    return _extract_whole_number_with_text_nl(tokens, short_scale, ordinals)


def _extract_fraction_with_text_nl(tokens, short_scale, ordinals):
    """Extract fraction numbers from a string.

    This function handles text such as '2 and 3/4'. Note that "one half" or
    similar will be parsed by the whole number function.

    Args:
        tokens [Token]: words and their indexes in the original string.
        short_scale boolean:
        ordinals boolean:

    Returns:
        (int or float, [Token])
        The value found, and the list of relevant tokens.
        (None, None) if no fraction value is found.
    """
    for c in _FRACTION_MARKER_NL:
        partitions = partition_list(tokens, lambda t: t.word == c)

        if len(partitions) == 3:
            numbers1 = \
                _extract_numbers_with_text_nl(partitions[0], short_scale,
                                              ordinals, fractional_numbers=False)
            numbers2 = \
                _extract_numbers_with_text_nl(partitions[2], short_scale,
                                              ordinals, fractional_numbers=True)

            if not numbers1 or not numbers2:
                return None, None

            # ensure first is not a fraction and second is a fraction
            num1 = numbers1[-1]
            num2 = numbers2[0]
            if num1.value >= 1 and 0 < num2.value < 1:
                return num1.value + num2.value, \
                       num1.tokens + partitions[1] + num2.tokens

    return None, None


def _extract_decimal_with_text_nl(tokens, short_scale, ordinals):
    """Extract decimal numbers from a string.

    This function handles text such as '2 point 5'.

    Notes:
        While this is a helper for extractnumber_nl, it also depends on
        extractnumber_nl, to parse out the components of the decimal.

        This does not currently handle things like:
            number dot number number number

    Args:
        tokens [Token]: The text to parse.
        short_scale boolean:
        ordinals boolean:

    Returns:
        (float, [Token])
        The value found and relevant tokens.
        (None, None) if no decimal value is found.
    """
    for c in _DECIMAL_MARKER_NL:
        partitions = partition_list(tokens, lambda t: t.word == c)

        if len(partitions) == 3:
            numbers1 = \
                _extract_numbers_with_text_nl(partitions[0], short_scale,
                                              ordinals, fractional_numbers=False)
            numbers2 = \
                _extract_numbers_with_text_nl(partitions[2], short_scale,
                                              ordinals, fractional_numbers=False)

            if not numbers1 or not numbers2:
                return None, None

            number = numbers1[-1]
            decimal = numbers2[0]

            # TODO handle number dot number number number
            if "." not in str(decimal.text):
                return number.value + float('0.' + str(decimal.value)), \
                       number.tokens + partitions[1] + decimal.tokens
    return None, None


def _extract_whole_number_with_text_nl(tokens, short_scale, ordinals):
    """Handle numbers not handled by the decimal or fraction functions.

    This is generally whole numbers. Note that phrases such as "one half" will
    be handled by this function, while "one and a half" are handled by the
    fraction function.

    Args:
        tokens [Token]:
        short_scale boolean:
        ordinals boolean:

    Returns:
        int or float, [_Tokens]
        The value parsed, and tokens that it corresponds to.
    """
    multiplies, string_num_ordinal, string_num_scale = \
        _initialize_number_data_nl(short_scale)

    number_words = []  # type: [Token]
    val = False
    prev_val = None
    next_val = None
    to_sum = []
    for idx, token in enumerate(tokens):
        current_val = None
        if next_val:
            next_val = None
            continue

        word = token.word
        if word in _ARTICLES_NL or word in _NEGATIVES_NL:
            number_words.append(token)
            continue

        prev_word = tokens[idx - 1].word if idx > 0 else ""
        next_word = tokens[idx + 1].word if idx + 1 < len(tokens) else ""

        if word not in string_num_scale and \
                word not in _STRING_NUM_NL and \
                word not in _SUMS_NL and \
                word not in multiplies and \
                not (ordinals and word in string_num_ordinal) and \
                not is_numeric(word) and \
                not is_fractional_nl(word, short_scale=short_scale) and \
                not look_for_fractions(word.split('/')):
            words_only = [token.word for token in number_words]
            if number_words and not all([w in _ARTICLES_NL |
                                         _NEGATIVES_NL for w in words_only]):
                break
            else:
                number_words = []
                continue
        elif word not in multiplies \
                and prev_word not in multiplies \
                and prev_word not in _SUMS_NL \
                and not (ordinals and prev_word in string_num_ordinal) \
                and prev_word not in _NEGATIVES_NL \
                and prev_word not in _ARTICLES_NL:
            number_words = [token]
        elif prev_word in _SUMS_NL and word in _SUMS_NL:
            number_words = [token]
        else:
            number_words.append(token)

        # is this word already a number ?
        if is_numeric(word):
            if word.isdigit():  # doesn't work with decimals
                val = int(word)
            else:
                val = float(word)
            current_val = val

        # is this word the name of a number ?
        if word in _STRING_NUM_NL:
            val = _STRING_NUM_NL.get(word)
            current_val = val
        elif word in string_num_scale:
            val = string_num_scale.get(word)
            current_val = val
        elif ordinals and word in string_num_ordinal:
            val = string_num_ordinal[word]
            current_val = val

        # is the prev word an ordinal number and current word is one?
        # second one, third one
        if ordinals and prev_word in string_num_ordinal and val == 1:
            val = prev_val

        # is the prev word a number and should we sum it?
        # twenty two, fifty six
        if prev_word in _SUMS_NL and val and val < 10:
            val = prev_val + val

        # is the prev word a number and should we multiply it?
        # twenty hundred, six hundred
        if word in multiplies:
            if not prev_val:
                prev_val = 1
            val = prev_val * val

        # is this a spoken fraction?
        # half cup
        if val is False:
            val = is_fractional_nl(word, short_scale=short_scale)
            current_val = val

        # 2 fifths
        if not ordinals:
            next_val = is_fractional_nl(next_word, short_scale=short_scale)
            if next_val:
                if not val:
                    val = 1
                val = val * next_val
                number_words.append(tokens[idx + 1])

        # is this a negative number?
        if val and prev_word and prev_word in _NEGATIVES_NL:
            val = 0 - val

        # let's make sure it isn't a fraction
        if not val:
            # look for fractions like "2/3"
            aPieces = word.split('/')
            if look_for_fractions(aPieces):
                val = float(aPieces[0]) / float(aPieces[1])
                current_val = val

        else:
            if prev_word in _SUMS_NL and word not in _SUMS_NL and current_val >= 10:
                # Backtrack - we've got numbers we can't sum.
                number_words.pop()
                val = prev_val
                break
            prev_val = val

            # handle long numbers
            # six hundred sixty six
            # two million five hundred thousand
            if word in multiplies and next_word not in multiplies:
                to_sum.append(val)
                val = 0
                prev_val = 0

    if val is not None and to_sum:
        val += sum(to_sum)

    return val, number_words


def _initialize_number_data_nl(short_scale):
    """Generate dictionaries of words to numbers, based on scale.

    This is a helper function for _extract_whole_number.

    Args:
        short_scale boolean:

    Returns:
        (set(str), dict(str, number), dict(str, number))
        multiplies, string_num_ordinal, string_num_scale
    """
    multiplies = _MULTIPLIES_SHORT_SCALE_NL if short_scale \
        else _MULTIPLIES_LONG_SCALE_NL

    string_num_ordinal_nl = _STRING_SHORT_ORDINAL_NL if short_scale \
        else _STRING_LONG_ORDINAL_NL

    string_num_scale_nl = _SHORT_SCALE_NL if short_scale else _LONG_SCALE_NL
    string_num_scale_nl = invert_dict(string_num_scale_nl)

    return multiplies, string_num_ordinal_nl, string_num_scale_nl


def extract_number_nl(text, short_scale=True, ordinals=False):
    """Extract a number from a text string

    The function handles pronunciations in long scale and short scale

    https://en.wikipedia.org/wiki/Names_of_large_numbers

    Args:
        text (str): the string to normalize
        short_scale (bool): use short scale if True, long scale if False
        ordinals (bool): consider ordinal numbers, third=3 instead of 1/3
    Returns:
        (int) or (float) or False: The extracted number or False if no number
                                   was found
    """
    return _extract_number_with_text_nl(tokenize(text.lower()),
                                        short_scale, ordinals).value


def is_fractional_nl(input_str, short_scale=True):
    """This function takes the given text and checks if it is a fraction.

    Args:
        input_str (str): the string to check if fractional
        short_scale (bool): use short scale if True, long scale if False
    Returns:
        (bool) or (float): False if not a fraction, otherwise the fraction
    """
    fracts = {"heel": 1, "half": 2, "halve": 2, "kwart": 4}
    if short_scale:
        for num in _SHORT_ORDINAL_STRING_NL:
            if num > 2:
                fracts[_SHORT_ORDINAL_STRING_NL[num]] = num
    else:
        for num in _LONG_ORDINAL_STRING_NL:
            if num > 2:
                fracts[_LONG_ORDINAL_STRING_NL[num]] = num

    if input_str.lower() in fracts:
        return 1.0 / fracts[input_str.lower()]
    return False
